#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel TARS model to generate charge by ionization."""

import subprocess
import typing as t  # noqa: F401
from bisect import bisect
from pathlib import Path

import numpy as np
import pandas as pd
from typing_extensions import Literal

from pyxel.detectors import Detector
from pyxel.models.charge_generation.tars.particle import Particle
from pyxel.models.charge_generation.tars.util import (
    load_histogram_data,
    read_data,
    sampling_distribution,
)


class Simulation:
    """Main class of the program, Simulation contain all the methods to set and run a simulation."""

    def __init__(self, detector: Detector) -> None:
        """Initialize the simulation.

        :param Detector detector: Detector object(from CCD/CMSO library) containing all the simulated detector specs
        """
        self.detector = detector
        self.simulation_mode = (
            None
        )  # type: t.Optional[Literal["cosmic_ray", "cosmics","radioactive_decay", "snowflakes"]]

        self.flux_dist = None  # type: t.Optional[np.ndarray]
        self.spectrum_cdf = None  # type: t.Optional[np.ndarray]

        self.energy_loss_data = (
            None
        )  # type: t.Optional[Literal['stopping', 'stepsize', 'geant4']]

        self.elec_number_dist = pd.DataFrame()
        self.elec_number_cdf = np.zeros((1, 2))
        self.step_size_dist = pd.DataFrame()
        self.step_cdf = np.zeros((1, 2))
        self.kin_energy_dist = pd.DataFrame()
        self.kin_energy_cdf = np.zeros((1, 2))

        self.data_library = pd.DataFrame()

        self.stopping_power = None  # type: t.Optional[np.ndarray]

        self.particle = None  # type: t.Optional[Particle]

        self.particle_type = (
            None
        )  # type: t.Optional[Literal["proton", "ion", "alpha", "beta", "electron", "gamma", "x-ray"]]
        self.initial_energy = (
            None
        )  # type: t.Optional[t.Union[int, float, Literal['random']]]
        self.position_ver = None  # type: t.Optional[str]
        self.position_hor = None  # type: t.Optional[str]
        self.position_z = None  # type: t.Optional[str]
        self.angle_alpha = None  # type: t.Optional[str]
        self.angle_beta = None  # type: t.Optional[str]
        self.step_length = (
            1.0  # fix, all the other data/parameters should be adjusted to this
        )
        self.energy_cut = 1.0e-5  # MeV

        self.e_num_lst_per_step = []  # type: t.List[int]
        self.e_energy_lst = []  # type: t.List[float]
        self.e_pos0_lst = []  # type: t.List[float]
        self.e_pos1_lst = []  # type: t.List[float]
        self.e_pos2_lst = []  # type: t.List[float]
        self.e_vel0_lst = np.array([])  # type: np.ndarray
        self.e_vel1_lst = np.array([])  # type: np.ndarray
        self.e_vel2_lst = np.array([])  # type: np.ndarray

        self.electron_number_from_eloss = []  # type: t.List[int]
        self.secondaries_from_eloss = []  # type: t.List[int]
        self.tertiaries_from_eloss = []  # type: t.List[int]

        self.track_length_lst_per_event = []  # type: t.List[float]
        self.e_num_lst_per_event = []  # type: t.List[int]
        self.sec_lst_per_event = []  # type: t.List[int]
        self.ter_lst_per_event = []  # type: t.List[int]
        self.edep_per_step = []  # type: t.List[float]
        self.total_edep_per_particle = []  # type: t.List[float]
        self.p_energy_lst_per_event = []  # type: t.List[float]
        self.alpha_lst_per_event = []  # type: t.List[float]
        self.beta_lst_per_event = []  # type: t.List[float]

    def parameters(
        self,
        sim_mode: Literal["cosmic_ray", "cosmics", "radioactive_decay", "snowflakes"],
        part_type: Literal[
            "proton", "ion", "alpha", "beta", "electron", "gamma", "x-ray"
        ],
        init_energy: t.Union[int, float, Literal["random"]],
        pos_ver: str,
        pos_hor: str,
        pos_z: str,
        alpha: str,
        beta: str,
    ) -> None:
        """TBW.

        :param sim_mode:
        :param part_type:
        :param init_energy:
        :param pos_ver:
        :param pos_hor:
        :param pos_z:
        :param alpha:
        :param beta:
        """
        self.simulation_mode = sim_mode
        self.particle_type = part_type
        self.initial_energy = init_energy
        self.position_ver = pos_ver
        self.position_hor = pos_hor
        self.position_z = pos_z
        self.angle_alpha = alpha
        self.angle_beta = beta

    def find_smaller_neighbor(self, column: str, value: float) -> float:
        """TBW.

        :return:
        """
        sorted_list = sorted(
            self.data_library[column].unique()
        )  # type: t.Sequence[float]
        index = bisect(sorted_list, value) - 1
        if index < 0:
            index = 0
        return sorted_list[index]

    def find_larger_neighbor(self, column: str, value: float) -> float:
        """TBW.

        :return:
        """
        sorted_list = sorted(
            self.data_library[column].unique()
        )  # type: t.Sequence[float]
        index = bisect(sorted_list, value)
        if index > len(sorted_list) - 1:
            index = len(sorted_list) - 1
        return sorted_list[index]

    def find_closest_neighbor(self, column: str, value: float) -> float:
        """TBW.

        :return:
        """
        sorted_list = sorted(
            self.data_library[column].unique()
        )  # type: t.Sequence[float]
        index_smaller = bisect(sorted_list, value) - 1
        index_larger = bisect(sorted_list, value)

        if index_larger >= len(sorted_list):
            return sorted_list[-1]
        elif (sorted_list[index_larger] - value) < (value - sorted_list[index_smaller]):
            return sorted_list[index_larger]
        else:
            return sorted_list[index_smaller]

    def select_stepsize_data(
        self, p_type: str, p_energy: float, p_track_length: float
    ) -> Path:
        """TBW.

        :param p_type: str
        :param p_energy: float (MeV)
        :param p_track_length: float (um)
        :return:
        """
        df = self.data_library

        distance = self.find_larger_neighbor(column="thickness", value=p_track_length)
        energy = self.find_closest_neighbor(column="energy", value=p_energy)

        df_filtered = df[
            (df.type == p_type) & (df.energy == energy) & (df.thickness == distance)
        ]  # type: pd.DataFrame

        serie = df_filtered["path"]  # type: pd.Series
        filename = serie.values[0]  # type: Path

        return filename

    def set_stepsize_distribution(self, step_size_file: Path) -> None:
        """TBW.

        :param step_size_file:
        :return:
        .. warning:: EXPERIMENTAL - NOT FINSHED YET
        """
        # # step size distribution in um
        self.step_size_dist = load_histogram_data(
            step_size_file, hist_type="step_size", skip_rows=4, read_rows=10000
        )

        cum_sum = np.cumsum(self.step_size_dist["counts"])
        cum_sum /= np.max(cum_sum)
        self.step_cdf = np.stack((self.step_size_dist["step_size"], cum_sum), axis=1)

        # # tertiary electron numbers created by secondary electrons
        self.elec_number_dist = load_histogram_data(
            step_size_file, hist_type="electron", skip_rows=10008, read_rows=10000
        )

        cum_sum_2 = np.cumsum(self.elec_number_dist["counts"])
        cum_sum_2 /= np.max(cum_sum_2)
        self.elec_number_cdf = np.stack(
            (self.elec_number_dist["electron"] - 0.5, cum_sum_2), axis=1
        )

        # # secondary electron spectrum in keV
        # self.kin_energy_dist = load_histogram_data(step_size_file, hist_type='energy', skip_rows=10008, read_rows=200)
        #
        # cum_sum = np.cumsum(self.kin_energy_dist['counts'])
        # cum_sum /= np.max(cum_sum)
        # self.kin_energy_cdf = np.stack((self.kin_energy_dist['energy'], cum_sum), axis=1)

    def event_generation(self) -> bool:
        """Generate an event.

        :return:
        """
        track_left = False
        electron_number_per_event = 0
        secondary_per_event = 0
        tertiary_per_event = 0
        geo = self.detector.geometry
        mat = self.detector.material
        ioniz_energy = mat.ionization_energy  # eV

        assert self.simulation_mode is not None
        assert self.particle_type is not None
        assert self.initial_energy is not None
        assert self.spectrum_cdf is not None
        assert self.position_ver is not None
        assert self.position_hor is not None
        assert self.position_z is not None

        self.particle = Particle(
            detector=self.detector,
            simulation_mode=self.simulation_mode,
            particle_type=self.particle_type,
            input_energy=self.initial_energy,
            spectrum_cdf=self.spectrum_cdf,
            starting_pos_ver=self.position_ver,
            starting_pos_hor=self.position_hor,
            starting_pos_z=self.position_z
            # self.angle_alpha, self.angle_beta)
        )

        particle = self.particle
        assert particle.track_length is not None

        self.track_length_lst_per_event += [particle.track_length]

        if self.energy_loss_data == "stepsize":
            # data_filename = self.select_stepsize_data(particle.type, particle.energy, particle.track_length)
            data_filename = self.select_stepsize_data(particle.type, 1000.0, 40.0)
            self.set_stepsize_distribution(data_filename)
            # TODO make a stack of stepsize cdfs and do not load them more than once!!!
        # elif self.energy_loss_data == 'geant4':
        #     pass
        elif self.energy_loss_data == "stopping":
            raise NotImplementedError  # TODO: implement this

        while True:
            if particle.energy <= self.energy_cut:
                break

            # particle.energy is in MeV !
            # particle.deposited_energy is in keV !

            if self.energy_loss_data == "stepsize":
                current_step_size = sampling_distribution(self.step_cdf)  # um
                # e_kin_energy = sampling_distribution(self.kin_energy_cdf)     # keV   TODO
            # elif self.energy_loss_data == 'geant4':
            #     pass
            elif self.energy_loss_data == "stopping":
                raise NotImplementedError  # TODO: implement this

            e_kin_energy = 1.0  # TODO
            particle.deposited_energy = e_kin_energy + ioniz_energy * 1e-3  # keV

            # UPDATE POSITION OF IONIZING PARTICLES
            particle.position[0] += particle.dir_ver * current_step_size  # um
            particle.position[1] += particle.dir_hor * current_step_size  # um
            particle.position[2] += particle.dir_z * current_step_size  # um

            # check if p is still inside detector and have enough energy:
            if (
                particle.position[0] <= 0.0
                or particle.position[0] >= geo.vert_dimension
            ):
                break
            if (
                particle.position[1] <= 0.0
                or particle.position[1] >= geo.horz_dimension
            ):
                break
            if (
                particle.position[2] <= -1 * geo.total_thickness
                or particle.position[2] >= 0.0
            ):
                break
            if particle.deposited_energy >= particle.energy * 1e3:
                break

            track_left = True

            particle.energy -= particle.deposited_energy * 1e-3  # MeV

            if self.energy_loss_data == "stepsize":
                # the +1 is the original secondary electron
                electron_number = int(sampling_distribution(self.elec_number_cdf)) + 1
                # electron_number = int(e_kin_energy * 1e3 / ioniz_energy) + 1
            # elif self.energy_loss_data == 'geant4':
            #     electron_number = electron_number_vector[g4_j]
            #     g4_j += 1
            elif self.energy_loss_data == "stopping":
                raise NotImplementedError

            secondary_per_event += 1
            tertiary_per_event += electron_number - 1
            electron_number_per_event += electron_number
            self.e_num_lst_per_step.append(electron_number)

            self.e_energy_lst.append(e_kin_energy * 1e3)  # eV
            self.e_pos0_lst.append(particle.position[0])  # um
            self.e_pos1_lst.append(particle.position[1])  # um
            self.e_pos2_lst.append(particle.position[2])  # um

            self.edep_per_step.append(particle.deposited_energy)  # keV
            particle.total_edep += particle.deposited_energy  # keV

            # save particle trajectory
            particle.trajectory = np.vstack((particle.trajectory, particle.position))
        # END of loop

        if track_left:
            self.total_edep_per_particle.append(particle.total_edep)  # keV
            self.e_num_lst_per_event.append(electron_number_per_event)
            self.sec_lst_per_event.append(secondary_per_event)
            self.ter_lst_per_event.append(tertiary_per_event)

        return False

    def event_generation_geant4(self) -> bool:
        """Generate an event running a geant4 app directly.

        :return:
        """
        # error = None
        electron_number_per_event = 0
        secondary_per_event = 0
        tertiary_per_event = 0

        secondaries = 0
        tertiaries = 0

        assert self.simulation_mode is not None
        assert self.particle_type is not None
        assert self.initial_energy is not None
        assert self.spectrum_cdf is not None
        assert self.position_ver is not None
        assert self.position_hor is not None
        assert self.position_z is not None

        self.particle = Particle(
            detector=self.detector,
            simulation_mode=self.simulation_mode,
            particle_type=self.particle_type,
            input_energy=self.initial_energy,
            spectrum_cdf=self.spectrum_cdf,
            starting_pos_ver=self.position_ver,
            starting_pos_hor=self.position_hor,
            starting_pos_z=self.position_z
            # self.angle_alpha, self.angle_beta
        )
        particle = self.particle

        assert particle.track_length is not None
        if particle.track_length < 1.0:
            return True

        self.track_length_lst_per_event.append(particle.track_length)
        self.p_energy_lst_per_event.append(particle.energy)
        self.alpha_lst_per_event.append(particle.alpha)
        self.beta_lst_per_event.append(particle.beta)

        error = subprocess.call(
            [
                "./pyxel/models/charge_generation/tars/data/geant4/TestEm18",
                "Silicon",
                particle.type,
                str(particle.energy),
                str(particle.track_length),
            ],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.PIPE,
        )
        if error != 0:
            return True

        # mat = self.detector.material
        # subprocess.call(['./TestEm18', mat.xxx, particle.type,
        # str(particle.energy), str(particle.track_length)'])

        g4_data_energy_path = Path(__file__).parent.joinpath(
            "data", "geant4", "tars_geant4_energy.data"
        )
        g4energydata = read_data(g4_data_energy_path)  # MeV

        # primary_e_balance = g4energydata[0] * 1.E6
        all_e_loss = g4energydata[1] * 1.0e6
        primary_e_loss = g4energydata[2] * 1.0e6
        secondary_e_loss = g4energydata[3] * 1.0e6
        if all_e_loss > 0.0:
            self.electron_number_from_eloss.append(
                np.floor(all_e_loss / 3.6).astype(int)
            )
            self.secondaries_from_eloss.append(
                np.floor(primary_e_loss / 3.6).astype(int)
            )
            self.tertiaries_from_eloss.append(
                np.floor(secondary_e_loss / 3.6).astype(int)
            )

        g4_data_path = Path(__file__).parent.joinpath(
            "data", "geant4", "tars_geant4.data"
        )
        g4data = read_data(g4_data_path)  # mm (!)

        if g4data.shape == (
            3,
        ):  # alternative running mode, only all electron number without proton step size data
            electron_number_vector = [g4data[0].astype(int)]
            secondaries = g4data[1].astype(int)
            tertiaries = g4data[2].astype(int)
            step_size_vector = [0]
        elif g4data.shape == (0,):
            step_size_vector = []  # um
            electron_number_vector = []
        elif g4data.shape == (2,):
            step_size_vector = [g4data[0] * 1e3]  # um
            electron_number_vector = [g4data[1].astype(int)]
        else:
            step_size_vector = g4data[:, 0] * 1e3  # um
            electron_number_vector = g4data[:, 1].astype(int)

        if np.any(electron_number_vector):
            # for j in range(len(step_size_vector)):
            for j in range(len(electron_number_vector)):

                # UPDATE POSITION OF IONIZING PARTICLES
                particle.position[0] += particle.dir_ver * step_size_vector[j]  # um
                particle.position[1] += particle.dir_hor * step_size_vector[j]  # um
                particle.position[2] += particle.dir_z * step_size_vector[j]  # um

                electron_number_per_event += electron_number_vector[j]
                secondary_per_event += secondaries
                tertiary_per_event += tertiaries

                self.e_num_lst_per_step.append(electron_number_vector[j])
                self.e_pos0_lst.append(particle.position[0])  # um
                self.e_pos1_lst.append(particle.position[1])  # um
                self.e_pos2_lst.append(particle.position[2])  # um

                e_kin_energy = 1.0
                self.e_energy_lst.append(e_kin_energy)  # eV

                # self.edep_per_step.append(particle.deposited_energy)    # keV
                # particle.total_edep += particle.deposited_energy        # keV

                # save particle trajectory
                particle.trajectory = np.vstack(
                    (particle.trajectory, particle.position)
                )

            # END of loop

            # self.total_edep_per_particle.append(particle.total_edep)  # keV
            self.e_num_lst_per_event.append(electron_number_per_event)
            self.sec_lst_per_event.append(secondary_per_event)
            self.ter_lst_per_event.append(tertiary_per_event)

        # print('p energy: ', particle.energy, '\ttrack length: ', particle.track_length,
        #       '\telectrons/event: ', electron_number_per_event,
        #       '\tsteps: ', len(step_size_vector), '\terror: ', error)

        return False

    # def _ionization_(self, particle):
    #     """TBW.
    #
    #     :param particle:
    #     :return:
    #     """
    #     geo = self.detector.geometry
    #     ioniz_energy = geo.ionization_energy
    #     let_value = None
    #
    #     # particle.energy is in MeV !
    #     # particle.deposited_energy is in keV !
    #     if self.energy_loss_data == 'let':
    #         let_value = sampling_distribution(self.let_cdf)  # keV/um
    #     elif self.energy_loss_data == 'stopping':
    #         stopping_power = get_yvalue_with_interpolation(self.stopping_power, particle.energy)  # MeV*cm2/g
    #         let_value = 0.1 * stopping_power * geo.material_density  # keV/um
    #
    #     particle.deposited_energy = let_value * self.step_length  # keV
    #
    #     if particle.deposited_energy >= particle.energy * 1e3:
    #         particle.deposited_energy = particle.energy * 1e3
    #
    #     e_kin_energy = 0.1  # eV
    #     electron_number = int(particle.deposited_energy * 1e3 / (ioniz_energy + e_kin_energy))  # eV/eV = 1
    #
    #     self.e_num_lst += [electron_number]
    #     self.e_energy_lst += [e_kin_energy]
    #     self.e_pos0_lst += [particle.position[0]]
    #     self.e_pos1_lst += [particle.position[1]]
    #     self.e_pos2_lst += [particle.position[2]]
    #
    #     # keV
    #     particle.deposited_energy = electron_number * (e_kin_energy + ioniz_energy) * 1e-3
    #     particle.energy -= particle.deposited_energy * 1e-3     # MeV
    #
    #     self.edep_per_step.append(particle.deposited_energy)    # keV
    #     particle.total_edep += particle.deposited_energy        # keV



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""
import typing as t
from collections import abc
from enum import Enum
from numbers import Number

from typing_extensions import Literal

from pyxel.evaluator import eval_range


class ParameterType(Enum):
    """TBW."""

    # one-dimensional, can be a dataset coordinate
    Simple = "simple"
    # multi-dimensional or loaded from a file in parallel mode, cannot be a dataset coordinate
    Multi = "multi"


class ParameterValues:
    """Contains keys and values of parameters in a parametric step."""

    def __init__(
        self,
        key: str,
        values: t.Union[
            Literal["_"], t.Sequence[Literal["_"]], t.Sequence[Number], t.Sequence[str]
        ],
        boundaries: t.Optional[t.Tuple[float, float]] = None,
        enabled: bool = True,
        logarithmic: bool = False,
    ):

        # TODO: maybe use numpy to check multi-dimensional input lists
        # Check YAML input (not real values yet) and define parameter type
        if values == "_":
            self.type = ParameterType("multi")
        elif isinstance(values, str) and "numpy" in values:
            self.type = ParameterType("simple")
        elif isinstance(values, abc.Sequence) and any(
            [(el == "_" or isinstance(el, abc.Sequence)) for el in values]
        ):
            self.type = ParameterType("multi")
        elif isinstance(values, abc.Sequence) and all(
            [isinstance(el, (Number, str)) for el in values]
        ):
            self.type = ParameterType("simple")
        else:
            raise ValueError("Parameter values cannot be initiated with those values.")

        if "sampling.start_time" in key:
            key = "detector.sampling_properties.start_time"

        # unique identifier to the step. example: 'detector.geometry.row'
        self._key = key  # type: str
        self._values = (
            values
        )  # type: t.Union[Literal["_"], t.Sequence[Literal["_"]], t.Sequence[Number], t.Sequence[str]]

        # short  name identifier: 'row'
        self._short_name = key.split(".")[-1]

        self._enabled = enabled  # type: bool
        self._logarithmic = logarithmic  # type: bool
        self._boundaries = boundaries  # type: t.Optional[t.Tuple[float, float]]

        self._current = None  # type: t.Optional[t.Union[Literal['_'], Number, str]]

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__  # type: str

        return f"{cls_name}<key={self.key!r}, values={self.values!r}, enabled={self.enabled!r}>"

    def __len__(self) -> int:
        values = eval_range(self.values)  # type: t.Sequence
        return len(values)

    # TODO: Is method '__contains__' needed ? If yes then this class will act as a `Collections.abc.Sequence`
    def __iter__(self) -> t.Iterator[Number]:
        values = eval_range(self.values)
        for value in values:
            yield value

    @property
    def key(self) -> str:
        """TBW."""
        return self._key

    @property
    def short_name(self) -> str:
        """TBW."""
        return self._short_name

    @property
    def values(
        self,
    ) -> t.Union[
        Literal["_"], t.Sequence[Literal["_"]], t.Sequence[Number], t.Sequence[str]
    ]:
        """TBW."""
        return self._values

    @property
    def enabled(self) -> bool:
        """TBW."""
        return self._enabled

    @property
    def current(self) -> t.Optional[t.Union[Literal["_"], Number, str]]:
        """TBW."""
        return self._current

    @current.setter
    def current(self, value: t.Union[Literal["_"], str, Number]) -> None:
        """TBW."""
        self._current = value

    @property
    def logarithmic(self) -> bool:
        """TBW."""
        return self._logarithmic

    @property
    def boundaries(self) -> t.Optional[t.Tuple[float, float]]:
        """TBW."""
        return self._boundaries



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""
import typing as t

from pyxel.pipelines import ModelFunction, ModelGroup


class DetectionPipeline:
    """TBW."""

    # TODO: develop a ModelGroupList class ? Pass this as a single argument.
    def __init__(
        self,  # TODO: Too many instance attributes
        photon_generation: t.Optional[ModelGroup] = None,
        optics: t.Optional[ModelGroup] = None,
        phasing: t.Optional[ModelGroup] = None,
        charge_generation: t.Optional[ModelGroup] = None,
        charge_collection: t.Optional[ModelGroup] = None,
        charge_transfer: t.Optional[ModelGroup] = None,
        charge_measurement: t.Optional[ModelGroup] = None,
        signal_transfer: t.Optional[ModelGroup] = None,
        readout_electronics: t.Optional[ModelGroup] = None,
        doc: t.Optional[str] = None,
    ):
        self._is_running = False
        self.doc = doc

        self.photon_generation = photon_generation  # type: t.Optional[ModelGroup]
        self.optics = optics  # type: t.Optional[ModelGroup]
        self.phasing = phasing  # type: t.Optional[ModelGroup] # MKID-array
        self.charge_generation = charge_generation  # type: t.Optional[ModelGroup]
        self.charge_collection = charge_collection  # type: t.Optional[ModelGroup]
        self.charge_measurement = charge_measurement  # type: t.Optional[ModelGroup]
        self.readout_electronics = readout_electronics  # type: t.Optional[ModelGroup]
        self.charge_transfer = charge_transfer  # type: t.Optional[ModelGroup]  # CCD
        self.signal_transfer = signal_transfer  # type: t.Optional[ModelGroup]  # CMOS

        # TODO: this defines the order of steps in the pipeline.
        #       The ModelGroupList could do this. Is it really needed?
        self.MODEL_GROUPS = (
            "photon_generation",
            "optics",
            "phasing",
            "charge_generation",
            "charge_collection",
            "charge_transfer",
            "charge_measurement",
            "signal_transfer",
            "readout_electronics",
        )  # type: t.Tuple[str, ...]

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__  # type: str
        return f"{cls_name}<is_running={self._is_running!r}, doc={self.doc!r}>"

    def __iter__(self) -> t.Iterable[ModelFunction]:
        for model in self.MODEL_GROUPS:
            models_grp = getattr(self, model)  # type: t.Optional[ModelGroup]
            if models_grp:
                yield from models_grp

    # def run(self, detector, abort_before: t.Optional[str]=None) -> None:
    #     """TBW."""
    #     self._is_running = True
    #     for model_func in iter(self):
    #         if model_func.name == abort_before:
    #             break
    #
    #         model_func(detector)
    #
    #     self._is_running = False

    # def __getstate__(self):
    #     """TBW."""
    #     return {
    #         'photon_generation': self.photon_generation,
    #         'optics': self.optics,
    #         'charge_generation': self.charge_generation,
    #         'charge_collection': self.charge_collection,
    #         'charge_transfer': self.charge_transfer,
    #         'charge_measurement': self.charge_measurement,
    #         'signal_transfer': self.signal_transfer,
    #         'readout_electronics': self.readout_electronics,
    #         '_model_groups': self.model_group_names,              # TODO
    #     }

    @property
    def model_group_names(self) -> t.Tuple[str, ...]:
        """TBW."""
        return self.MODEL_GROUPS

    @property
    def is_running(self) -> bool:
        """Return the running state of this pipeline."""
        return self._is_running

    def abort(self) -> None:
        """TBW."""
        self._is_running = False

    # TODO: Is this method used ?
    def get_model(self, name: str) -> ModelFunction:
        """Return a ModelFunction object for the specified model name.

        Parameters
        ----------
        name: str
            Name of the model.

        Returns
        -------
        None

        Raises
        ------
        AttributeError
            If model with the specified name is not found.
        """
        for group_name in self.model_group_names:  # type: str
            model_group = getattr(self, group_name)  # type: ModelGroup
            if model_group:
                for model in model_group.models:  # type: ModelFunction
                    if name == model.name:
                        return model
        raise AttributeError("Model has not been found.")



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Simple model to convert photon into photo-electrons inside detector."""
import logging

import numpy as np

from pyxel.detectors import Detector


# TODO: Fix this
# @validators.validate
# @config.argument(name='', label='', units='', validate=)
def simple_conversion(detector: Detector) -> None:
    """Generate charge from incident photon via photoelectric effect, simple statistical model.

    Parameters
    ----------
    detector : Detector
        Pyxel Detector object.
    """
    geo = detector.geometry
    ch = detector.characteristics
    ph = detector.photon

    detector_charge = np.zeros(
        (geo.row, geo.col)
    )  # all pixels has zero charge by default
    photon_rows, photon_cols = ph.array.shape
    detector_charge[slice(0, photon_rows), slice(0, photon_cols)] = (
        ph.array * ch.qe * ch.eta
    )
    charge_number = detector_charge.flatten()  # the average charge numbers per pixel
    where_non_zero = np.where(charge_number > 0.0)
    charge_number = charge_number[where_non_zero]
    size = charge_number.size

    init_ver_pix_position = geo.vertical_pixel_center_pos_list()[where_non_zero]
    init_hor_pix_position = geo.horizontal_pixel_center_pos_list()[where_non_zero]

    detector.charge.add_charge(
        particle_type="e",
        particles_per_cluster=charge_number,
        init_energy=np.zeros(size),
        init_ver_position=init_ver_pix_position,
        init_hor_position=init_hor_pix_position,
        init_z_position=np.zeros(size),
        init_ver_velocity=np.zeros(size),
        init_hor_velocity=np.zeros(size),
        init_z_velocity=np.zeros(size),
    )


# TODO: Fix this
# @validators.validate
# @config.argument(name='', label='', units='', validate=)
def monte_carlo_conversion(detector: Detector) -> None:
    """Generate charge from incident photon via photoelectric effect, more exact, stochastic (Monte Carlo) model.

    :param detector: Pyxel Detector object
    """
    logging.info("")

    # detector.qe <= 1
    # detector.eta <= 1
    # if np.random.rand(size) <= detector.qe:
    #     pass    # 1 e
    # else:
    #     pass
    # if np.random.rand(size) <= detector.eta:
    #     pass    # 1 e
    # else:
    #     pass
    # TODO: random number for QE
    # TODO: random number for eta
    # TODO: energy threshold


def random_pos(detector: Detector) -> None:
    """Generate random position for photoelectric effect inside detector.

    :param detector: Pyxel Detector object
    """
    # pos1 = detector.vert_dimension * np.random.random()
    # pos2 = detector.horz_dimension * np.random.random()

    # size = 0
    # pos3 = -1 * detector.total_thickness * np.random.rand(size)
    # return pos3
    raise NotImplementedError



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Readout electronics model."""
import logging

from pyxel.detectors import Detector

# from astropy import units as u


# @pyxel.validate
# @pyxel.argument(name='', label='', units='', validate=)
def simple_amplifier(detector: Detector) -> None:
    """Amplify signal.

    amp - Gain of output amplifier, a1 - Gain of the signal processor

    :param detector: Pyxel Detector object
    """
    logging.info("")
    char = detector.characteristics
    detector.signal.array *= char.amp * char.a1



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Readout noise model."""
import logging
import typing as t

import numpy as np

from pyxel.detectors import Detector

# from astropy import units as u


# @validators.validate
# @config.argument(name='', label='', units='', validate=)


def output_node_noise(
    detector: Detector, std_deviation: float, random_seed: t.Optional[int] = None
) -> None:
    """Adding noise to signal array of detector output node using normal random distribution.

    detector Signal unit: Volt

    :param detector: Pyxel Detector object
    :param std_deviation: standard deviation
    :param random_seed: seed
    """
    logging.info("")
    if random_seed:
        np.random.seed(random_seed)

    signal_mean_array = detector.signal.array.astype("float64")
    sigma_array = std_deviation * np.ones(signal_mean_array.shape)

    signal = np.random.normal(loc=signal_mean_array, scale=sigma_array)

    detector.signal.array = signal



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel charge collection model."""
import logging

import numba
import numpy as np

from pyxel.data_structure import Charge
from pyxel.detectors import Detector


# @pyxel.validate
# @pyxel.argument(name='', label='', units='', validate=)
# TODO: Is it not better/relevant if all models return a new `Detector` ?
#       Therefore the 'input' `Detector` is never modified.
#
#       Example:
#       >>> def simple_collection(detector: Detector) -> Detector:
#       ...     # Create a new object `Detector`.
#       ...     # Note: This is not an expensive 'deep' copy
#       ...     new_detector = detector.copy()
#       ...     # Do something
#       ...     return new_detector
def simple_collection(detector: Detector) -> None:
    """Associate charge with the closest pixel."""
    logging.info("")
    geo = detector.geometry
    array = np.zeros((geo.row, geo.col))

    charge_per_pixel = detector.charge.get_values(quantity="number")
    charge_pos_ver = detector.charge.get_values(quantity="position_ver")
    charge_pos_hor = detector.charge.get_values(quantity="position_hor")

    pixel_index_ver = np.floor_divide(charge_pos_ver, geo.pixel_vert_size).astype(int)
    pixel_index_hor = np.floor_divide(charge_pos_hor, geo.pixel_horz_size).astype(int)

    # Changing = to += since charge dataframe is reset, the pixel array need to be
    # incremented, we can't do the whole operation on each iteration
    detector.pixel.array += df_to_array(
        array, charge_per_pixel, pixel_index_ver, pixel_index_hor
    ).astype(np.int32)


def empty_charge(detector: Detector) -> None:
    """Each time the charges are collected in the pixel, the charge array is reset using Charge().

    This allows to limit memory leaks due to long exposure.
    There will still be a problem for very large charge array due to very high flux
    in simulation.
    """
    detector._charge = Charge()


# TODO: Is it needed not better to first create a copy of 'array' and then work with this copy ??
@numba.jit(nopython=True)
def df_to_array(
    array: np.ndarray,
    charge_per_pixel: list,
    pixel_index_ver: list,
    pixel_index_hor: list,
) -> np.ndarray:
    """TBW."""
    for i, charge_value in enumerate(charge_per_pixel):
        array[pixel_index_ver[i], pixel_index_hor[i]] += charge_value
    return array



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
"""TBW."""

import typing as t

import numpy as np

from pyxel.evaluator import eval_range
from pyxel.inputs.loader import load_table


class Readout:
    """TBW."""

    def __init__(
        self,
        times: t.Optional[t.Union[t.Sequence, str]] = None,
        times_from_file: t.Optional[str] = None,
        start_time: float = 0.0,
        non_destructive: bool = False,
    ):
        """Create an instance of Readout class.

        Parameters
        ----------
        times
        times_from_file
        start_time
        non_destructive
        """
        self._time_domain_simulation = True

        if times is not None and times_from_file is not None:
            raise ValueError("Both times and times_from_file specified. Choose one.")
        elif times is None and times_from_file is None:
            self._times = np.array(
                [1]
            )  # by convention default readout/exposure time is 1 second
            self._time_domain_simulation = False
        elif times_from_file:
            self._times = load_table(times_from_file).to_numpy(dtype=float).flatten()
        elif times:
            self._times = np.array(eval_range(times), dtype=float)
        else:
            raise ValueError("Sampling times not specified.")

        if self._times[0] == 0:
            raise ValueError("Readout times should be non-zero values.")
        elif start_time == self._times[0]:
            raise ValueError("Readout times should be greater than start time.")

        self._non_destructive = non_destructive

        self._times_linear = True  # type: bool
        self._start_time = start_time  # type:float
        self._steps = np.array([])  # type: np.ndarray
        self._num_steps = 0  # type: int
        self._set_steps()

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__  # type: str
        return f"{cls_name}<num_steps={self._num_steps}>"

    def _set_steps(self) -> None:
        """TBW."""
        self._times, self._steps = calculate_steps(self._times, self._start_time)
        self._times_linear = bool(np.all(self._steps == self._steps[0]))
        self._num_steps = len(self._times)

    def time_step_it(self) -> t.Iterator[t.Tuple[float, float]]:
        """TBW."""
        return zip(self._times, self._steps)

    @property
    def times(self) -> t.Any:
        """TBW."""
        return self._times

    @property
    def time_domain_simulation(self) -> bool:
        """TBW."""
        return self._time_domain_simulation

    @property
    def steps(self) -> np.ndarray:
        """TBW."""
        return self._steps

    @property
    def non_destructive(self) -> bool:
        """TBW."""
        return self._non_destructive


def calculate_steps(
    times: np.ndarray, start_time: float
) -> t.Tuple[np.ndarray, np.ndarray]:
    """Calculate time differences for a given array and start time.

    Parameters
    ----------
    times: ndarray
    start_time: float

    Returns
    -------
    times: ndarray
        Modified times according to start time.
    steps: ndarray
        Steps corresponding to times.
    """
    steps = np.diff(
        np.concatenate((np.array([start_time]), times), axis=0),
        axis=0,
    )

    return times, steps



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""
from pyxel.detectors import Characteristics


# TODO: 'svg' should be the full volume and not the half
class CCDCharacteristics(Characteristics):
    """Characteristical attributes of a CCD detector."""

    def __init__(
        self,
        # Parameters for `Characteristics`
        qe: float = 0.0,  # unit: NA
        eta: float = 0.0,  # unit: electron/photon
        sv: float = 0.0,  # unit: volt/electron
        amp: float = 0.0,  # unit: V/V
        a1: float = 0.0,  # unit: V/V
        a2: int = 0,  # unit: adu/V
        fwc: int = 0,  # unit: electron
        vg: float = 0.0,  # unit: cm^2
        dt: float = 0.0,  # unit: s
        # Parameters specific `CCDCharacteristics`
        fwc_serial: int = 0,  # unit: electron
        svg: float = 0.0,  # unit: cm^2
        t: float = 0.0,  # unit: s
        st: float = 0.0,  # unit: s
    ):
        """Create an instance of `CCDCharacteristics`.

        Parameters
        ----------
        fwc_serial: int
            Full well capacity (serial register). Unit: electron
        svg: float
            Half pixel volume charge can occupy (serial register). Unit: cm^2
        t: float
            Parallel transfer period. Unit: s
        st: float
            Serial transfer period. Unit: s
        """
        if fwc_serial not in range(10_000_001):
            raise ValueError("'fwc_serial' must be between 0 and 1e+7.")

        if not (0.0 <= svg <= 1.0):
            raise ValueError("'svg' must be between 0.0 and 1.0.")

        if not (0.0 <= t <= 10.0):
            raise ValueError("'t' must be between 0.0 and 10.0.")

        if not (0.0 <= st <= 10.0):
            raise ValueError("'st' must be between 0.0 and 10.0.")

        super().__init__(
            qe=qe, eta=eta, sv=sv, amp=amp, a1=a1, a2=a2, fwc=fwc, vg=vg, dt=dt
        )
        self._fwc_serial = fwc_serial
        self._svg = svg
        self._t = t
        self._st = st

    def __repr__(self) -> str:
        return f"CCDCharacteristics(fwc_serial={self.fwc_serial}, svg={self.svg}, t={self.t}, st={self.st})"

    @property
    def fwc_serial(self) -> int:
        """Get Full well capacity (serial register)."""
        return self._fwc_serial

    @fwc_serial.setter
    def fwc_serial(self, value: int) -> None:
        """Set Full well capacity (serial register)."""
        if value not in range(10_000_001):
            raise ValueError("'fwc_serial' must be between 0 and 1e+7.")

        self._fwc_serial = value

    @property
    def svg(self) -> float:
        """Get Half pixel volume charge can occupy (serial register)."""
        return self._svg

    @svg.setter
    def svg(self, value: float) -> None:
        """Set Half pixel volume charge can occupy (serial register)."""
        if not (0.0 <= value <= 1.0):
            raise ValueError("'svg' must be between 0.0 and 1.0.")

        self._svg = value

    @property
    def t(self) -> float:
        """Get Parallel transfer period."""
        return self._t

    @t.setter
    def t(self, value: float) -> None:
        """Set Parallel transfer period."""
        if not (0.0 <= value <= 10.0):
            raise ValueError("'t' must be between 0.0 and 10.0.")

        self._t = value

    @property
    def st(self) -> float:
        """Set Serial transfer period."""
        return self._st

    @st.setter
    def st(self, value: float) -> None:
        """Get Serial transfer period."""
        if not (0.0 <= value <= 10.0):
            raise ValueError("'st' must be between 0.0 and 10.0.")

        self._st = value



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel TARS model to generate charge by ionization."""

import typing as t
from pathlib import Path

import numpy as np
import pandas as pd

if t.TYPE_CHECKING:
    from pyxel.models.charge_generation.tars.tars import TARS

try:
    from matplotlib import pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D  # noqa: F401
except ImportError:
    # raise Warning('Matplotlib cannot be imported')
    pass


class PlottingTARS:
    """Plotting class for TARS."""

    def __init__(
        self,
        tars: "TARS",
        draw_plots: bool = False,
        save_plots: bool = False,
        file_format: str = "png",
    ) -> None:
        self.tars = tars

        self.draw_plots = draw_plots
        self.save_plots = save_plots
        self.file_format = file_format

    def save_and_draw(self, fig_name: str) -> None:
        """TBW.

        :param fig_name:
        """
        file_name = fig_name + "." + self.file_format
        if self.save_plots:
            plt.savefig(file_name)
        if self.draw_plots:
            plt.draw()

    def show(self) -> None:
        """TBW."""
        plt.show()

    def save_edep(self) -> None:
        """TBW."""
        np.save("orig2_edep_per_step_10k", self.tars.sim_obj.edep_per_step)
        np.save(
            "orig2_edep_per_particle_10k", self.tars.sim_obj.total_edep_per_particle
        )

    def plot_edep_per_step(self) -> tuple:
        """TBW."""
        plt.figure()
        n, bins, patches = plt.hist(self.tars.sim_obj.edep_per_step, 300, facecolor="b")
        plt.xlabel("E_dep (keV)")
        plt.ylabel("Counts")
        plt.title("Histogram of E deposited per step")
        # plt.axis([0, 0.003, 0, 1.05*max(n)])
        plt.grid(True)
        self.save_and_draw("edep_per_step")
        return n, bins, patches

    def plot_edep_per_particle(self) -> tuple:
        """TBW."""
        plt.figure()
        n, bins, patches = plt.hist(
            self.tars.sim_obj.total_edep_per_particle, 200, facecolor="g"
        )
        plt.xlabel("E_dep (keV)")
        plt.ylabel("Counts")
        plt.title("Histogram of total E deposited per particle")
        # plt.axis([0, 0.4, 0, 1.05*max(n)])
        plt.grid(True)
        self.save_and_draw("edep_per_particle")
        return n, bins, patches

    def plot_spectrum_cdf(self) -> None:
        """TBW."""
        sim_obj = self.tars.sim_obj
        assert sim_obj.spectrum_cdf is not None
        assert sim_obj.flux_dist is not None

        lin_energy_range = sim_obj.spectrum_cdf[:, 0]
        cum_sum = sim_obj.spectrum_cdf[:, 1]
        plt.figure()
        plt.title("Spectrum CDF")
        plt.xlabel("Energy (MeV)")
        plt.ylabel("Probability")
        plt.semilogx(lin_energy_range, cum_sum)
        self.save_and_draw("spectrum_cdf")

    def plot_flux_spectrum(self) -> None:
        """TBW."""
        sim_obj = self.tars.sim_obj
        assert sim_obj.spectrum_cdf is not None
        assert sim_obj.flux_dist is not None

        lin_energy_range = sim_obj.spectrum_cdf[:, 0]
        flux_dist = sim_obj.flux_dist
        plt.figure()
        plt.title("Proton flux spectrum (CREME data)")
        plt.xlabel("Energy (MeV)")
        plt.ylabel("Flux (1/(s*MeV))")
        plt.loglog(lin_energy_range, flux_dist)
        self.save_and_draw("flux_spectrum")

    def plot_spectrum_hist(self, data: t.Optional[str] = None) -> None:
        """TBW."""
        plt.figure()
        plt.title("Proton flux spectrum sampled by TARS")
        plt.xlabel("Energy (MeV)")
        plt.ylabel("Counts")
        # plt.ylabel('Flux (1/(s*MeV))')
        # plt.loglog(lin_energy_range, flux_dist)

        if isinstance(data, str):
            if data.endswith(".npy"):
                data = np.load(data)

        hist_bins = 250
        # hist_range = (1e-1, 1e5)
        # col = (0, 1, 1, 1)
        plt.hist(data, bins=np.logspace(np.log10(0.1), np.log10(1e5), hist_bins))
        plt.gca().set_xscale("log")
        # plt.legend(loc='upper right')
        self.save_and_draw("tars_spectrum")

    def plot_charges_3d(self) -> None:
        """TBW."""
        geo = self.tars.sim_obj.detector.geometry

        # set up a figure twice as wide as it is tall
        fig = plt.figure(figsize=plt.figaspect(0.5))
        # fig = plt.figure()

        e_num_lst_per_event = self.tars.sim_obj.e_num_lst_per_event
        size = [x / 10.0 for x in e_num_lst_per_event]
        # ax = fig.add_subplot(1, 2, 1, projection='3d')
        ax = fig.add_subplot(1, 2, 1, projection="3d")
        ax.scatter(
            self.tars.sim_obj.e_pos0_lst,
            self.tars.sim_obj.e_pos1_lst,
            self.tars.sim_obj.e_pos2_lst,
            c="b",
            marker=".",
            s=size,
        )
        # ax.plot(self.tars.sim_obj.particle.trajectory[:, 0],
        #         self.tars.sim_obj.particle.trajectory[:, 1],
        #         self.tars.sim_obj.particle.trajectory[:, 2], 'c-')

        # ax2 = fig.add_subplot(1, 2, 2, projection='3d')
        # ax2.scatter(self.tars.sim_obj.e_pos0_lst, self.tars.sim_obj.e_pos1_lst, 0,
        #             c='r', marker='.', s=size)

        # ax.hold(True)
        point1 = np.array([0, 0, 0])
        point2 = np.array([0, 0, -1 * geo.total_thickness])
        # point3 = np.array([1, 2, 3])
        normal = np.array([0, 0, 1])
        # norma2 = np.array([0, 1, 0])
        # norma3 = np.array([1, 0, 0])

        # point2 = np.array([10, 50, 50])

        # a plane is a*x+b*y+c*z+d=0
        # [a,b,c] is the normal. Thus, we have to calculate
        # d and we're set
        d1 = -point1.dot(normal)
        d2 = -point2.dot(normal)
        # d3 = -point3.dot(norma3)

        # create x,y
        xx, yy = np.meshgrid(
            range(int(geo.vert_dimension)), range(int(geo.horz_dimension))
        )

        # calculate corresponding z
        z1 = (-normal[0] * xx - normal[1] * yy - d1) * 1.0 / normal[2]
        z2 = (-normal[0] * xx - normal[1] * yy - d2) * 1.0 / normal[2]
        # z3 = (-norma3[0] * xx - norma3[1] * yy - d3) * 1. / norma3[2]

        ax.plot_surface(xx, yy, z1, alpha=0.2, color=(0, 0, 1))
        ax.plot_surface(xx, yy, z2, alpha=0.2, color=(0, 0, 1))
        # ax.plot_surface(xx, yy, z3, alpha=0.2)

        ax.set_xlim(0, geo.vert_dimension)
        ax.set_ylim(0, geo.horz_dimension)
        ax.set_zlim(-1 * geo.total_thickness, 0)
        ax.set_xlabel(r"vertical ($\mu$m)")
        ax.set_ylabel(r"horizontal ($\mu$m)")
        ax.set_zlabel(r"z ($\mu$m)")

        # ax2.set_xlim(0, geo.vert_dimension)
        # ax2.set_ylim(0, geo.horz_dimension)
        # ax2.set_zlim(-1 * geo.total_thickness, 0)
        # ax2.set_xlabel(r'vertical ($\mu$m)')
        # ax2.set_ylabel(r'horizontal ($\mu$m)')
        # ax2.set_zlabel(r'z ($\mu$m)')

    # def plot_let_cdf(self) -> None:
    #     """TBW."""
    #     plt.figure()
    #     plt.title('LET CDF')
    #     plt.plot(self.tars.sim_obj.let_cdf[:, 0], self.tars.sim_obj.let_cdf[:, 1], '.')
    #     self.save_and_draw('let_cdf')

    def plot_step_cdf(self) -> None:
        """TBW."""
        plt.figure()
        plt.title("Step size CDF")
        plt.plot(
            self.tars.sim_obj.step_cdf[:, 0], self.tars.sim_obj.step_cdf[:, 1], "."
        )
        self.save_and_draw("step_cdf")

    def plot_step_dist(self) -> None:
        """TBW."""
        plt.figure()
        plt.title("Step size distribution")
        plt.xlabel("step size (um)")
        plt.ylabel("counts")
        plt.plot(
            self.tars.sim_obj.step_size_dist["step_size"],
            self.tars.sim_obj.step_size_dist["counts"],
            ".",
        )
        self.save_and_draw("step_dist")

    def plot_tertiary_number_cdf(self) -> None:
        """TBW."""
        plt.figure()
        plt.title("Tertiary electron number CDF")
        plt.plot(
            self.tars.sim_obj.elec_number_cdf[:, 0],
            self.tars.sim_obj.elec_number_cdf[:, 1],
            ".",
        )
        self.save_and_draw("elec_number_cdf")

    def plot_tertiary_number_dist(self) -> None:
        """TBW."""
        plt.figure()
        plt.title("Tertiary electron number distribution")
        plt.plot(
            self.tars.sim_obj.elec_number_dist["electron"],
            self.tars.sim_obj.elec_number_dist["counts"],
            ".",
        )
        self.save_and_draw("elec_number_dist")

    # def plot_let_dist(self) -> None:
    #     """TBW."""
    #     plt.figure()
    #     plt.title('LET dist')
    #     plt.plot(self.tars.sim_obj.let_dist[:, 1], self.tars.sim_obj.let_dist[:, 2], '.')
    #     self.save_and_draw('let_dist')

    def plot_trajectory_xy(self) -> None:
        """TBW."""
        plt.figure()
        geo = self.tars.sim_obj.detector.geometry

        particle = self.tars.sim_obj.particle
        assert particle is not None

        # self.trajectory[:, 0] - VERTICAL COORDINATE
        # self.trajectory[:, 1] - HORIZONTAL COORDINATE
        plt.plot(particle.trajectory[:, 1], particle.trajectory[:, 0], ".")
        plt.xlabel(r"horizontal ($\mu$m)")
        plt.ylabel(r"vertical ($\mu$m)")
        plt.title("p trajectory in CCD")
        plt.axis([0, geo.horz_dimension, 0, geo.vert_dimension])
        plt.grid(True)
        self.save_and_draw("trajectory_xy")

    def plot_trajectory_xz(self) -> None:
        """TBW."""
        plt.figure()
        geo = self.tars.sim_obj.detector.geometry
        particle = self.tars.sim_obj.particle
        assert particle is not None

        # self.trajectory[:, 2] - Z COORDINATE
        # self.trajectory[:, 1] - HORIZONTAL COORDINATE
        plt.plot(particle.trajectory[:, 1], particle.trajectory[:, 2], ".")
        plt.xlabel(r"horizontal ($\mu$m)")
        plt.ylabel(r"z ($\mu$m)")
        plt.title("p trajectory in CCD")
        plt.axis([0, geo.horz_dimension, -1 * geo.total_thickness, 0])
        plt.grid(True)
        self.save_and_draw("trajectory_xz")

    def plot_track_histogram(
        self, histogram_data: t.Union[str, np.ndarray], normalize: bool = False
    ) -> None:
        """TBW."""
        hist_bins = 1000
        hist_range = (0, 1000)

        plt.figure()
        plt.title("Proton track length distribution")

        if isinstance(histogram_data, str):
            if histogram_data.endswith(".npy"):
                data = np.load(histogram_data)  # type: np.ndarray
            else:
                raise NotImplementedError
        else:
            data = histogram_data

        col = (1, 0, 1, 1)

        if normalize:
            plt.hist(data, bins=hist_bins, range=hist_range, fc=col, density=True)
        else:
            plt.hist(data, bins=hist_bins, range=hist_range, fc=col)

        # plt.axis([0, 15e3, 0, 0.0001])
        # plt.axis([0, 15e3, 0, 3E3])

        plt.xlabel("Track length")
        plt.ylabel("Counts")
        plt.legend(loc="upper right")
        self.save_and_draw("track_length")

    def plot_step_size_histograms(self, normalize: bool = False) -> None:
        """TBW."""
        energies = ["100MeV"]
        thicknesses = ["40um", "50um", "60um", "70um", "100um"]
        p_types = ["proton"]

        path = Path(__file__).parent.joinpath("data", "inputs")

        # step_rows = 10000

        plt.figure()
        plt.title("Step size")
        for p_type in p_types:
            for energy in energies:
                for thickness in thicknesses:
                    filename = (
                        "stepsize_"
                        + p_type
                        + "_"
                        + energy
                        + "_"
                        + thickness
                        + "_Si_10k.ascii"
                    )
                    step_size = pd.read_csv(
                        str(Path(path, filename)),
                        delimiter="\t",
                        names=["step_size", "counts"],
                        usecols=[1, 2],
                        skiprows=4,
                        nrows=10000,
                    )

                    if normalize:
                        plotted_counts = step_size["counts"] / sum(step_size["counts"])
                    else:
                        plotted_counts = step_size["counts"]

                    plt.plot(
                        step_size["step_size"],
                        plotted_counts,
                        ".-",
                        label=p_type + ", " + energy + ", " + thickness,
                    )

        plt.axis([0, 20, 0, 0.010])
        plt.xlabel("Step size (um)")
        plt.ylabel("Counts")
        plt.legend(loc="upper right")
        self.save_and_draw("step_size_histograms")

    #
    # def plot_secondary_spectra(self, normalize: bool=None):
    #     """TBW."""
    #     energies = ['100MeV', '1GeV']
    #     thicknesses = ['10um', '50um', '100um', '200um']
    #     p_types = ['proton']
    #
    #     path = Path(__file__).parent.joinpath('data', 'inputs')
    #
    #     # step_rows = 10000
    #
    #     plt.figure()
    #     plt.title('Electron spectrum')
    #     for p_type in p_types:
    #         for energy in energies:
    #             for thickness in thicknesses:
    #                 filename = 'stepsize_' + p_type + '_' + energy + '_' + thickness + '_1M.ascii'
    #                 spectrum = pd.read_csv(str(Path(path, filename)),
    #                                        delimiter="\t", names=["energy", "counts"], usecols=[1, 2],
    #                                        skiprows=10008, nrows=200)
    #
    #                 if normalize:
    #                     plotted_counts = spectrum['counts'] / sum(spectrum['counts'])
    #                 else:
    #                     plotted_counts = spectrum['counts']
    #
    #                 plt.plot(spectrum['energy'], plotted_counts, '.-',
    #                          label=p_type + ', ' + energy + ', ' + thickness)
    #
    #     # plt.axis([0, 6.0, 0, 0.12])
    #     plt.xlabel('Energy')
    #     plt.ylabel('Counts')
    #     plt.legend(loc='upper right')
    #     self.save_and_draw('secondary_spectra')

    def plot_gaia_vs_gras_hist(self, normalize: bool = False) -> None:
        """TBW."""
        # GRAS simulation results (by Marco) + GAIA BAM data - Perfect overlap in case of normalization!
        path = Path(__file__).parent.joinpath(
            "data",
            "validation",
            "Gaia_CCD_study-20180404T115340Z-001",
            "Gaia_CCD_study",
            "Data",
        )
        hist_names = [
            "CRs_from_BAM_Gaia_CCDs.npy",  # GAIA BAM data
            "complete_G4_H_He_GCR_sim_deposition.npy",  # G4, contains a lot of events with ZERO number of e-!
            r"C:\dev\work\pyxel\pyxel\models\charge_generation"
            + r"\tars\data\validation\G4_app_results_20180425\tars-e_num_lst_per_event.npy",
        ]
        labels = ["Gaia BAM CCD data", "GRAS simulation", "TARS (Pyxel) simulation"]
        i = 0

        hist_bins = 250
        hist_range = (1, 15e3)

        plt.figure()
        ax = plt.axes()

        plt.title("Charges deposited per single event")
        for filename in hist_names:

            histogram = np.load(str(Path(path, filename)))

            if i == 0:
                col = (0, 0, 1, 0.7)  # blue
            if i == 1:
                col = (0, 1, 0, 0.7)  # green
            if i == 2:
                col = (1, 0, 0, 0.7)  # red
            # cyan      (0, 1, 1, 0.5)
            # magenta   (1, 0, 1, 0.5)

            if normalize:
                plt.hist(
                    histogram,
                    bins=hist_bins,
                    range=hist_range,
                    histtype="step",
                    label=labels[i],
                    color=col,
                    density=True,
                )
            else:
                plt.hist(
                    histogram,
                    bins=hist_bins,
                    range=hist_range,
                    histtype="step",
                    label=labels[i],
                    color=col,
                )

            i += 1

        plt.axis([0, 15e3, 0, 3.0e-4])

        plt.xlabel("Number of electrons")
        plt.ylabel("Counts (normalized)")

        ax.ticklabel_format(axis="y", style="sci", scilimits=(0, 0))
        plt.legend(loc="upper right")
        self.save_and_draw("gaia_vs_gras_electron_hist")

    def plot_old_tars_hist(self, normalize: bool = False) -> None:
        """TBW."""
        # earlier TARS results of Lionel
        folder_path = Path(__file__).parent.joinpath(
            "data", "validation", "Results-20180404T121902Z-001", "Results"
        )
        hist_names = [
            # '10000 events from random protons CREME96 (16um - SM)(22-08-2016_16h36)',  # 16 um SM
            "10000 events from random protons CREME96 (16um - SM)(22-08-2016_16h41)",  # 16 um SM
            "10000 events from random protons CREME96 (step=0.5)(16-08-2016_15h56)",  # 40 um BAM
        ]
        labels = ["TARS data (Lionel), SM (16um)", "TARS data (Lionel), BAM (40um)"]
        i = 0

        hist_bins = 500
        hist_range = (1, 15e3)

        plt.figure()
        plt.title("Number of electrons per event")
        for filename in hist_names:

            histogram = np.load(
                str(
                    Path(
                        folder_path.joinpath(filename),
                        "Raw data/Electrons_generated.npy",
                    )
                )
            )

            if i == 0:
                col = (1, 0, 0, 0.5)
            if i == 1:
                col = (0, 1, 0, 0.5)

            if normalize:
                plt.hist(
                    histogram,
                    bins=hist_bins,
                    range=hist_range,
                    label=labels[i],
                    fc=col,
                    density=True,
                )
            else:
                plt.hist(
                    histogram, bins=hist_bins, range=hist_range, label=labels[i], fc=col
                )

            i += 1

        # plt.axis([0, 15e3, 0, 0.0001])
        # plt.axis([0, 15e3, 0, 3E3])

        plt.xlabel("")
        plt.ylabel("Counts")
        plt.legend(loc="upper right")
        self.save_and_draw("old_tars_electron_hist")

    def plot_gaia_bam_vs_sm_electron_hist(self, normalize: bool = False) -> None:
        """TBW."""
        path = Path(__file__).parent.joinpath("data", "validation")
        hist_names = [
            # 'Gaia_bam_ccd_events(13259).npy', NEM JOOOOO
            "Gaia_CCD_study-20180404T115340Z-001/Gaia_CCD_study/Data/CRs_from_SM_Gaia_CCDs.npy",
            "Gaia_CCD_study-20180404T115340Z-001/Gaia_CCD_study/Data/CRs_from_BAM_Gaia_CCDs.npy",
        ]
        labels = ["GAIA SM (16um) data", "GAIA BAM (40um) data"]
        i = 0

        hist_bins = 500
        hist_range = (1, 15e3)

        plt.figure()
        plt.title("Number of electrons per event")
        for filename in hist_names:

            histogram = np.load(str(Path(path, filename)))

            if i == 0:
                col = (1, 0, 0, 0.5)
            if i == 1:
                col = (0, 0, 1, 0.5)

            if normalize:
                plt.hist(
                    histogram,
                    bins=hist_bins,
                    range=hist_range,
                    label=labels[i],
                    fc=col,
                    density=True,
                )
            else:
                plt.hist(
                    histogram, bins=hist_bins, range=hist_range, label=labels[i], fc=col
                )

            i += 1

        # plt.axis([0, 15e3, 0, 0.0001])
        # plt.axis([0, 15e3, 0, 3E3])

        plt.xlabel("")
        plt.ylabel("Counts")
        plt.legend(loc="upper right")
        self.save_and_draw("gaia_BAM_vs_SM_electron_hist")

    def plot_electron_hist(
        self,
        data1: t.Union[str, np.ndarray],
        data2: t.Optional[np.ndarray] = None,
        data3: t.Optional[np.ndarray] = None,
        title: str = "",
        hist_bins: int = 500,
        hist_range: t.Tuple[int, int] = (0, 15000),
        normalize: bool = False,
    ) -> None:
        """TBW."""
        labels = [
            "TARS (David), 40um"
            # 'secondary e-',
            # 'tertiary e-'
        ]
        i = 0

        # hist_bins = 5000
        # hist_range = (0, 15000)

        # hist_bins = 500
        # hist_range = (10, 2E4)

        plt.figure()
        plt.title(title)

        if isinstance(data1, str):
            if data1.endswith(".npy"):
                data1 = np.load(data1)

        # data1 = data1[data1 > ]
        # data1 = data1[data1 < 15000]

        if normalize:
            plt.hist(
                data1,
                bins=hist_bins,
                range=hist_range,
                label=labels[i],
                fc=(1, 0, 0, 0.5),
                density=True,
            )
            if data2:
                i += 1
                plt.hist(
                    data2,
                    bins=hist_bins,
                    range=hist_range,
                    label=labels[i],
                    fc=(1, 1, 0, 0.5),
                    density=True,
                )
                if data3:
                    i += 1
                    plt.hist(
                        data3,
                        bins=hist_bins,
                        range=hist_range,
                        label=labels[i],
                        fc=(0, 1, 1, 0.5),
                        density=True,
                    )

        else:
            plt.hist(
                data1,
                bins=hist_bins,
                range=hist_range,
                label=labels[i],
                fc=(1, 0, 0, 0.5),
            )
            if data2:
                i += 1
                plt.hist(
                    data2,
                    bins=hist_bins,
                    range=hist_range,
                    label=labels[i],
                    fc=(1, 1, 0, 0.5),
                )
                if data3:
                    i += 1
                    plt.hist(
                        data3,
                        bins=hist_bins,
                        range=hist_range,
                        label=labels[i],
                        fc=(0, 1, 1, 0.5),
                    )

        # plt.axis([0, 15e3, 0, 0.0001])
        # plt.axis([0, 15e3, 0, 3E3])

        plt.xlabel("")
        plt.ylabel("Counts")
        plt.legend(loc="upper right")
        self.save_and_draw("electron_hist")

    def polar_angle_dist(self, theta: t.Union[str, np.ndarray]) -> None:
        """TBW."""
        if isinstance(theta, str):
            if theta.endswith(".npy"):
                theta_data = np.load(theta)  # type: np.ndarray
            else:
                raise NotImplementedError
        else:
            theta_data = theta

        fig = plt.figure()
        fig.add_subplot(111, polar=True)
        # theta = 2 * np.pi * np.random.rand(10000)
        plt.hist(theta_data, bins=360, histtype="step")
        # plt.polar(theta)
        plt.xlabel("")
        plt.ylabel("")
        plt.title("Incident angle distribution")
        self.save_and_draw("polar_angle")



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel TARS model to generate charge by ionization."""

import logging
import math
import typing as t  # noqa: F401
from pathlib import Path

import numpy as np
import pandas as pd
from tqdm.auto import tqdm
from typing_extensions import Literal

from pyxel.detectors import Detector
from pyxel.models.charge_generation.tars.plotting import PlottingTARS
from pyxel.models.charge_generation.tars.simulation import Simulation
from pyxel.models.charge_generation.tars.util import (  # , load_histogram_data
    interpolate_data,
    read_data,
)

# from astropy import units as u


# @validators.validate
# @config.argument(name='', label='', units='', validate=)
def run_tars(
    detector: Detector,
    simulation_mode: t.Optional[
        Literal["cosmic_ray", "cosmics", "radioactive_decay", "snowflakes"]
    ] = None,
    running_mode: t.Optional[
        Literal["stopping", "stepsize", "geant4", "plotting"]
    ] = None,
    particle_type: t.Optional[Literal["proton", "alpha", "ion"]] = None,
    initial_energy: t.Optional[t.Union[int, float, Literal["random"]]] = None,
    particle_number: t.Optional[int] = None,
    incident_angles: t.Optional[t.Tuple[str, str]] = None,
    starting_position: t.Optional[t.Tuple[str, str, str]] = None,
    # step_size_file: str = None,
    # stopping_file: str = None,
    spectrum_file: t.Optional[str] = None,
    random_seed: t.Optional[int] = None,
) -> None:
    """Simulate charge deposition by cosmic rays.

    :param detector: Pyxel detector object
    :param particle_type: type of particle: ``proton``, ``alpha``, ``ion``
    :param initial_energy: Kinetic energy of particle
    :param particle_number: Number of particles
    :param incident_angles: incident angles: ``(α, β)``
    :param starting_position: starting position: ``(x, y, z)``
    :param simulation_mode: simulation mode: ``cosmic_rays``, ``radioactive_decay``
    :param running_mode: mode: ``stopping``, ``stepsize``, ``geant4``, ``plotting``
    :param spectrum_file: path to input spectrum
    :param random_seed: seed
    """
    logging.info("")
    if random_seed:
        np.random.seed(random_seed)

    if simulation_mode is None:
        raise ValueError("TARS: Simulation mode is not defined")
    if running_mode is None:
        raise ValueError("TARS: Running mode is not defined")
    if particle_type is None:
        raise ValueError("TARS: Particle type is not defined")
    if particle_number is None:
        raise ValueError("TARS: Particle number is not defined")
    if spectrum_file is None:
        raise ValueError("TARS: Spectrum is not defined")

    if initial_energy is None:
        initial_energy = "random"  # TODO

    if incident_angles is None:
        incident_angle_alpha, incident_angle_beta = ("random", "random")
    else:
        incident_angle_alpha, incident_angle_beta = incident_angles

    if starting_position is None:
        start_pos_ver, start_pos_hor, start_pos_z = ("random", "random", "random")
    else:
        start_pos_ver, start_pos_hor, start_pos_z = starting_position

    tars = TARS(
        detector=detector,
        simulation_mode=simulation_mode,
        particle_type=particle_type,
        initial_energy=initial_energy,
        particle_number=particle_number,
        incident_angle_alpha=incident_angle_alpha,
        incident_angle_beta=incident_angle_beta,
        start_pos_ver=start_pos_ver,
        start_pos_hor=start_pos_hor,
        start_pos_z=start_pos_z,
    )

    # tars.set_simulation_mode(simulation_mode)
    # tars.set_particle_type(particle_type)                # MeV
    # tars.set_initial_energy(initial_energy)              # MeV
    # tars.set_particle_number(particle_number)            # -
    # tars.set_incident_angles(incident_angles)            # rad
    # tars.set_starting_position(starting_position)        # um
    tars.set_particle_spectrum(Path(spectrum_file))

    if running_mode == "stopping":
        # tars.run_mod()          ########
        raise NotImplementedError
        # tars.set_stopping_power(stopping_file)
        # tars.run()
    elif running_mode == "stepsize":
        tars.set_stepsize()
        tars.run()
    elif running_mode == "geant4":
        tars.set_geant4()
        tars.run()
    elif running_mode == "plotting":

        plot_obj = PlottingTARS(tars, save_plots=True, draw_plots=True)

        # # # plot_obj.plot_flux_spectrum()

        #
        # # plot_obj.plot_step_dist()
        # # plot_obj.plot_step_cdf()

        # plot_obj.plot_tertiary_number_cdf()
        # plot_obj.plot_tertiary_number_dist()

        # plot_obj.plot_step_size_histograms(normalize=True)
        # plot_obj.plot_secondary_spectra(normalize=True)
        #
        # # plot_obj.plot_edep_per_step()
        # # plot_obj.plot_edep_per_particle()

        # plot_obj.plot_charges_3d()

        plot_obj.plot_flux_spectrum()

        # plot_obj.plot_gaia_bam_vs_sm_electron_hist(normalize=True)
        # plot_obj.plot_old_tars_hist(normalize=True)

        plot_obj.plot_gaia_vs_gras_hist(normalize=True)

        # plot_obj.plot_track_histogram(tars.sim_obj.track_length_list)
        # plot_obj.plot_track_histogram(
        #     r'C:\dev\work\pyxel\pyxel\models\charge_generation\tars\data\validation\G4_app_results_20180425\tars-track_length_lst_per_event.npy',
        #     normalize=True)

        # plot_obj.plot_electron_hist(tars.sim_obj.e_num_lst_per_event, normalize=True)

        # plot_obj.plot_electron_hist(r'C:\dev\work\pyxel\pyxel\models\charge_generation\tars\data\validation\G4_app_results_20180420_2\tars-e_num_lst_per_step.npy',
        # plot_obj.plot_electron_hist(r'C:\dev\work\pyxel\pyxel\models\charge_generation\tars\data\validation\G4_app_results_20180420_2\tars-p_energy_lst_per_event.npy',

        # plot_obj.plot_electron_hist(r'C:\dev\work\pyxel\pyxel\models\charge_generation\tars\data\validation\G4_app_results_20180425\tars-e_num_lst_per_event.npy',
        #                             title='all e per event', hist_bins=500, hist_range=(0, 15000))

        # plot_obj.plot_electron_hist(r'C:\dev\work\pyxel\pyxel\models\charge_generation\tars\data\validation\G4_app_results_20180425\tars-sec_lst_per_event.npy',
        #                             title='secondary e per event', hist_bins=500, hist_range=(0, 15000))
        #
        # plot_obj.plot_electron_hist(r'C:\dev\work\pyxel\pyxel\models\charge_generation\tars\data\validation\G4_app_results_20180425\tars-ter_lst_per_event.npy',
        #                             title='tertiary e per event', hist_bins=500, hist_range=(0, 15000))

        # plot_obj.plot_spectrum_hist(
        #     r'C:\dev\work\pyxel\pyxel\models\charge_generation\tars\data\validation\G4_app_results_20180420_6\tars-p_energy_lst_per_event.npy')
        # plot_obj.plot_spectrum_hist(r'C:\dev\work\pyxel\tars-p_energy_lst_per_event.npy')

        # plot_obj.plot_electron_hist(r'C:\dev\work\pyxel\pyxel\models\charge_generation\tars\data\validation\G4_app_results_20180420\tars-e_num_lst_per_event.npy',
        #                             title='all e per event', hist_bins=500, hist_range=(0, 15000))
        #
        # plot_obj.plot_electron_hist(r'C:\dev\work\pyxel\pyxel\models\charge_generation\tars\data\validation\G4_app_results_20180420\tars-sec_lst_per_event.npy',
        #                             title='secondary e per event', hist_bins=400, hist_range=(0, 2000))
        #
        # plot_obj.plot_electron_hist(r'C:\dev\work\pyxel\pyxel\models\charge_generation\tars\data\validation\G4_app_results_20180420\tars-ter_lst_per_event.npy',
        #                             title='tertiary e per event', hist_bins=500, hist_range=(0, 5000))

        # plot_obj.plot_spectrum_hist(r'C:\dev\work\pyxel\pyxel\models\charge_generation\tars\data\validation\G4_app_results_20180420\tars-p_energy_lst_per_event.npy')

        # plot_obj.polar_angle_dist(r'C:\dev\work\pyxel\pyxel\models\charge_generation\tars\data\validation\G4_app_results_20180420_6\tars-alpha_lst_per_event.npy')
        # plot_obj.polar_angle_dist(r'C:\dev\work\pyxel\tars-alpha_lst_per_event.npy')

        # plot_obj.polar_angle_dist(r'C:\dev\work\pyxel\pyxel\models\charge_generation\tars\data\validation\G4_app_results_20180420_6\tars-beta_lst_per_event.npy')

        # plot_obj.polar_angle_dist(r'C:\dev\work\pyxel\tars-beta_lst_per_event.npy')
        # plot_obj.polar_angle_dist(
        #     r'C:\dev\work\pyxel\pyxel\models\charge_generation\tars\data\validation\Results-20180404T121902Z-001\
        # Results\All primary protons from Geant4 Gaia H He GCR(16-08-2016_11h18)\Raw data\alpha.npy')
        # plot_obj.polar_angle_dist(
        #     r'C:\dev\work\pyxel\pyxel\models\charge_generation\tars\data\validation\Results-20180404T121902Z-001\
        # Results\All primary protons from Geant4 Gaia H He GCR(16-08-2016_11h18)\Raw data\beta.npy')

        # plot_obj.polar_angle_dist(
        #     r'C:\dev\work\pyxel\pyxel\models\charge_generation\tars\data\validation\Results-20180404T121902Z-001\
        # Results\All primary protons from Geant4 Gaia H He GCR(16-08-2016_11h18)(17-08-2016_13h51)\Raw data\alpha.npy')
        # plot_obj.polar_angle_dist(
        #     r'C:\dev\work\pyxel\pyxel\models\charge_generation\tars\data\validation\Results-20180404T121902Z-001\
        # Results\All primary protons from Geant4 Gaia H He GCR(16-08-2016_11h18)(17-08-2016_13h51)\Raw data\beta.npy')

        # plot_obj.polar_angle_dist(
        #     r'C:\dev\work\pyxel\pyxel/models/charge_generation/tars/data/validation/Results-20180404T121902Z-001/
        # Results/10000 events from random protons CREME96 (step=0.5)(16-08-2016_15h56)\Raw data\alpha.npy')
        # plot_obj.polar_angle_dist(
        #     r'C:\dev\work\pyxel\pyxel\models\charge_generation\tars\data\validation\Results-20180404T121902Z-001\
        # Results\10000 events from random protons CREME96 (step=0.5)(16-08-2016_15h56)\Raw data\beta.npy')

        # todo: not implemented yet:
        # file_path = Path(__file__).parent.joinpath('data', 'inputs', 'all_elec_num_proton.ascii')
        # g4_all_e_num_hist = load_histogram_data(file_path, hist_type='electron', skip_rows=4, read_rows=1000)
        # plot_obj.plot_electron_hist(tars.sim_obj.e_num_lst_per_event, g4_all_e_num_hist, normalize=True)

        # plot_obj.plot_electron_hist(tars.sim_obj.e_num_lst_per_event,
        #                             tars.sim_obj.sec_lst_per_event,
        #                             tars.sim_obj.ter_lst_per_event, normalize=True)

        plot_obj.show()
    else:
        raise ValueError
    #
    # np.save('tars-e_num_lst_per_event.npy', tars.sim_obj.e_num_lst_per_event)
    # np.save('tars-sec_lst_per_event.npy', tars.sim_obj.sec_lst_per_event)
    # np.save('tars-ter_lst_per_event.npy', tars.sim_obj.ter_lst_per_event)
    # np.save('tars-track_length_lst_per_event.npy', tars.sim_obj.track_length_lst_per_event)
    # np.save('tars-p_energy_lst_per_event.npy', tars.sim_obj.p_energy_lst_per_event)
    # np.save('tars-alpha_lst_per_event.npy', tars.sim_obj.alpha_lst_per_event)
    # np.save('tars-beta_lst_per_event.npy', tars.sim_obj.beta_lst_per_event)
    # np.save('tars-e_num_lst_per_step.npy', tars.sim_obj.e_num_lst_per_step)

    # plot_obj = PlottingTARS(tars, save_plots=True, draw_plots=True)
    # plot_obj.plot_charges_3d()
    # plot_obj.show()


class TARS:
    """TBW."""

    def __init__(
        self,
        detector: Detector,
        simulation_mode: Literal[
            "cosmic_ray", "cosmics", "radioactive_decay", "snowflakes"
        ],
        particle_type: Literal[
            "proton", "ion", "alpha", "beta", "electron", "gamma", "x-ray"
        ],
        initial_energy: t.Union[int, float, Literal["random"]],
        particle_number: int,
        incident_angle_alpha: str,
        incident_angle_beta: str,
        start_pos_ver: str,
        start_pos_hor: str,
        start_pos_z: str,
    ):
        self.simulation_mode = simulation_mode
        self.part_type = particle_type
        self.init_energy = initial_energy
        self.particle_number = particle_number
        self.angle_alpha = incident_angle_alpha
        self.angle_beta = incident_angle_beta
        self.position_ver = start_pos_ver
        self.position_hor = start_pos_hor
        self.position_z = start_pos_z

        self.sim_obj = Simulation(detector)
        self.charge_obj = detector.charge
        self._log = logging.getLogger(__name__)

    # TODO: Is it still used ?
    def set_simulation_mode(
        self,
        sim_mode: Literal["cosmic_ray", "cosmics", "radioactive_decay", "snowflakes"],
    ) -> None:
        """TBW.

        :param sim_mode:
        """
        self.simulation_mode = sim_mode

    # TODO: Is it still used ?
    def set_particle_type(
        self,
        particle_type: Literal[
            "proton", "ion", "alpha", "beta", "electron", "gamma", "x-ray"
        ],
    ) -> None:
        """TBW.

        :param particle_type:
        """
        self.part_type = particle_type

    # TODO: Is it still used ?
    def set_initial_energy(
        self, energy: t.Union[int, float, Literal["random"]]
    ) -> None:
        """TBW.

        :param energy:
        """
        self.init_energy = energy

    # TODO: Is it still used ?
    def set_particle_number(self, number: int) -> None:
        """TBW.

        :param number:
        """
        self.particle_number = number

    # TODO: Is it still used ?
    def set_incident_angles(self, angles: t.Tuple[str, str]) -> None:
        """TBW.

        :param angles:
        """
        alpha, beta = angles
        self.angle_alpha = alpha
        self.angle_beta = beta

    # TODO: Is it still used ?
    def set_starting_position(self, start_position: t.Tuple[str, str, str]) -> None:
        """TBW.

        :param start_position:
        """
        position_vertical, position_horizontal, position_z = start_position
        self.position_ver = position_vertical
        self.position_hor = position_horizontal
        self.position_z = position_z

    def set_particle_spectrum(self, file_name: Path) -> None:
        """Set up the particle specs according to a spectrum.

        :param string file_name: path of the file containing the spectrum
        """
        spectrum = read_data(file_name)  # nuc/m2*s*sr*MeV
        geo = self.sim_obj.detector.geometry
        detector_area = geo.vert_dimension * geo.horz_dimension * 1.0e-8  # cm2

        spectrum[:, 1] *= 4 * math.pi * 1.0e-4 * detector_area  # nuc/s*MeV

        spectrum_function = interpolate_data(spectrum)

        lin_energy_range = np.arange(
            np.min(spectrum[:, 0]), np.max(spectrum[:, 0]), 0.01
        )
        self.sim_obj.flux_dist = spectrum_function(lin_energy_range)

        cum_sum = np.cumsum(self.sim_obj.flux_dist)
        cum_sum /= np.max(cum_sum)
        self.sim_obj.spectrum_cdf = np.stack((lin_energy_range, cum_sum), axis=1)

    def set_stopping_power(self, stopping_file: Path) -> None:
        """TBW.

        :param stopping_file:
        """
        self.sim_obj.energy_loss_data = "stopping"
        self.sim_obj.stopping_power = read_data(stopping_file)

    def set_stepsize(self) -> None:
        """TBW."""
        self.sim_obj.energy_loss_data = "stepsize"
        self.create_data_library()

    def set_geant4(self) -> None:
        """TBW."""
        self.sim_obj.energy_loss_data = "geant4"

    def create_data_library(self) -> None:
        """TBW."""
        self.sim_obj.data_library = pd.DataFrame(
            columns=["type", "energy", "thickness", "path"]
        )

        # mat_list = ['Si']

        type_list = [
            "proton"
        ]  # , 'ion', 'alpha', 'beta', 'electron', 'gamma', 'x-ray']
        energy_list = [100.0]  # MeV
        thick_list = [40.0, 50.0, 60.0, 70.0, 100.0]  # um

        # TODO: Fix this. See issue #152
        path = Path(__file__).parent.joinpath("data", "inputs")
        filename_list = [
            "stepsize_proton_100MeV_40um_Si_10k.ascii",
            "stepsize_proton_100MeV_50um_Si_10k.ascii",
            "stepsize_proton_100MeV_60um_Si_10k.ascii",
            "stepsize_proton_100MeV_70um_Si_10k.ascii",
            "stepsize_proton_100MeV_100um_Si_10k.ascii",
        ]

        i = 0
        for pt in type_list:
            for en in energy_list:
                for th in thick_list:
                    data_dict = {
                        "type": pt,
                        "energy": en,
                        "thickness": th,
                        "path": str(Path(path, filename_list[i])),
                    }
                    new_df = pd.DataFrame(data_dict, index=[0])
                    self.sim_obj.data_library = pd.concat(
                        [self.sim_obj.data_library, new_df], ignore_index=True
                    )
                    i += 1

    def run(self) -> None:
        """TBW."""
        # print("TARS - simulation processing...\n")

        self.sim_obj.parameters(
            sim_mode=self.simulation_mode,
            part_type=self.part_type,
            init_energy=self.init_energy,
            pos_ver=self.position_ver,
            pos_hor=self.position_hor,
            pos_z=self.position_z,
            alpha=self.angle_alpha,
            beta=self.angle_beta,
        )

        # Get output folder and create it (if needed)
        out_path = Path("data").resolve()
        out_path.mkdir(parents=True, exist_ok=True)

        self._log.info("Save data in folder '%s'", out_path)

        for k in tqdm(range(self.particle_number), desc="TARS", unit=" particle"):
            # for k in range(0, self.particle_number):
            err = None  # type: t.Optional[bool]
            if self.sim_obj.energy_loss_data == "stepsize":  # TODO
                err = self.sim_obj.event_generation()
            elif self.sim_obj.energy_loss_data == "geant4":
                err = self.sim_obj.event_generation_geant4()
            if k % 10 == 0:
                np.save(
                    f"{out_path}/tars-e_num_lst_per_event.npy",
                    self.sim_obj.e_num_lst_per_event,
                )
                np.save(
                    f"{out_path}/tars-sec_lst_per_event.npy",
                    self.sim_obj.sec_lst_per_event,
                )
                np.save(
                    f"{out_path}/tars-ter_lst_per_event.npy",
                    self.sim_obj.ter_lst_per_event,
                )
                np.save(
                    f"{out_path}/tars-track_length_lst_per_event.npy",
                    self.sim_obj.track_length_lst_per_event,
                )
                np.save(
                    f"{out_path}/tars-p_energy_lst_per_event.npy",
                    self.sim_obj.p_energy_lst_per_event,
                )
                np.save(
                    f"{out_path}/tars-alpha_lst_per_event.npy",
                    self.sim_obj.alpha_lst_per_event,
                )
                np.save(
                    f"{out_path}/tars-beta_lst_per_event.npy",
                    self.sim_obj.beta_lst_per_event,
                )

                np.save(
                    f"{out_path}/tars-e_num_lst_per_step.npy",
                    self.sim_obj.e_num_lst_per_step,
                )
                np.save(f"{out_path}/tars-e_pos0_lst.npy", self.sim_obj.e_pos0_lst)
                np.save(f"{out_path}/tars-e_pos1_lst.npy", self.sim_obj.e_pos1_lst)
                np.save(f"{out_path}/tars-e_pos2_lst.npy", self.sim_obj.e_pos2_lst)

                np.save(
                    f"{out_path}/tars-all_e_from_eloss.npy",
                    self.sim_obj.electron_number_from_eloss,
                )
                np.save(
                    f"{out_path}/tars-sec_e_from_eloss.npy",
                    self.sim_obj.secondaries_from_eloss,
                )
                np.save(
                    f"{out_path}/tars-ter_e_from_eloss.npy",
                    self.sim_obj.tertiaries_from_eloss,
                )
            if err:
                k -= 1

        size = len(self.sim_obj.e_num_lst_per_step)

        self.sim_obj.e_vel0_lst = np.zeros(size)
        self.sim_obj.e_vel1_lst = np.zeros(size)
        self.sim_obj.e_vel2_lst = np.zeros(size)

        self.charge_obj.add_charge(
            particle_type="e",
            particles_per_cluster=np.asarray(self.sim_obj.e_num_lst_per_step),
            init_energy=np.asarray(self.sim_obj.e_energy_lst),
            init_ver_position=np.asarray(self.sim_obj.e_pos0_lst),
            init_hor_position=np.asarray(self.sim_obj.e_pos1_lst),
            init_z_position=np.asarray(self.sim_obj.e_pos2_lst),
            init_ver_velocity=self.sim_obj.e_vel0_lst,
            init_hor_velocity=self.sim_obj.e_vel1_lst,
            init_z_velocity=self.sim_obj.e_vel2_lst,
        )

    def run_mod(self) -> None:
        """TBW."""
        # TODO: Use `logging`
        print("TARS - adding previous cosmic ray signals to image ...\n")

        # TODO: Use `pathlib.Path`
        out_path = "data/"
        e_num_lst_per_step = np.load(out_path + "tars-e_num_lst_per_step.npy")
        e_pos0_lst = np.load(out_path + "tars-e_pos0_lst.npy")
        e_pos1_lst = np.load(out_path + "tars-e_pos1_lst.npy")
        e_pos2_lst = np.load(out_path + "tars-e_pos2_lst.npy")

        size = len(e_num_lst_per_step)
        e_energy_lst = np.zeros(size)
        e_vel0_lst = np.zeros(size)
        e_vel1_lst = np.zeros(size)
        e_vel2_lst = np.zeros(size)

        self.charge_obj.add_charge(
            particle_type="e",
            particles_per_cluster=e_num_lst_per_step,
            init_energy=e_energy_lst,
            init_ver_position=e_pos0_lst,
            init_hor_position=e_pos1_lst,
            init_z_position=e_pos2_lst,
            init_ver_velocity=e_vel0_lst,
            init_hor_velocity=e_vel1_lst,
            init_z_velocity=e_vel2_lst,
        )



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""CMOS detector modeling class."""
import typing as t

from pyxel.detectors import Detector

if t.TYPE_CHECKING:
    from pyxel.detectors import CMOSCharacteristics, CMOSGeometry, Environment, Material


class CMOS(Detector):
    """CMOS-based detector class containing all detector attributes and data."""

    def __init__(
        self,
        geometry: "CMOSGeometry",
        material: "Material",
        environment: "Environment",
        characteristics: "CMOSCharacteristics",
    ):
        self._geometry = geometry  # type: CMOSGeometry
        self._characteristics = characteristics  # type: CMOSCharacteristics

        super().__init__(material=material, environment=environment)
        super().reset()

    @property
    def geometry(self) -> "CMOSGeometry":
        """TBW."""
        return self._geometry

    @property
    def characteristics(self) -> "CMOSCharacteristics":
        """TBW."""
        return self._characteristics



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2021.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

import os
from pathlib import Path

from freezegun import freeze_time

from pyxel.configuration import load


def test_load_2_times():
    """Test function 'pyxel.inputs_outputs.load' called two times."""
    filename = "tests/data/dummy_simple.yaml"

    # Get full filename
    full_filename = Path(filename)  # type: Path
    assert full_filename.exists()

    # Load the configuration file for the first time
    with freeze_time("2021-06-15 14:11"):
        _ = load(full_filename)

    # Load the configuration file for the second time
    with freeze_time("2021-06-15 14:11"):
        _ = load(full_filename)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""
from pyxel.detectors import Characteristics


class CMOSCharacteristics(Characteristics):
    """Characteristical attributes of a CMOS-based detector."""

    def __init__(
        self,
        # Parameters for `Characteristics`
        qe: float = 0.0,
        eta: float = 0.0,
        sv: float = 0.0,
        amp: float = 0.0,
        a1: float = 0.0,
        a2: int = 0,
        fwc: int = 0,
        vg: float = 0.0,
        dt: float = 0.0,
        # Parameters specific `CMOSCharacteristics`
        cutoff: float = 2.5,  # unit: um
        vbiaspower: float = 3.35,  # unit: V
        dsub: float = 0.5,  # unit: V
        vreset: float = 0.25,  # unit: V
        biasgate: float = 2.3,  # unit: V
        preampref: float = 1.7,  # unit: V
    ):
        """Create an instance of `CMOSCharacteristics`.

        Parameters
        ----------
        cutoff: float
            Cutoff wavelength. Unit: um
        vbiaspower: float
            VBIASPOWER. Unit: V
        dsub: float
            DSUB. Unit: V
        vreset: float
            VRESET. Unit: V
        biasgate: float
            BIASGATE. Unit: V
        preampref: float
            PREAMPREF. Unit: V
        """
        if not (1.7 <= cutoff <= 15.0):
            raise ValueError("'cutoff' must be between 1.7 and 15.0.")

        if not (0.0 <= vbiaspower <= 3.4):
            raise ValueError("'vbiaspower' must be between 0.0 and 3.4.")

        if not (0.3 <= dsub <= 1.0):
            raise ValueError("'dsub' must be between 0.3 and 1.0.")

        if not (0.0 <= vreset <= 0.3):
            raise ValueError("'vreset' must be between 0.0 and 0.3.")

        if not (1.8 <= biasgate <= 2.6):
            raise ValueError("'biasgate' must be between 1.8 and 2.6.")

        if not (0.0 <= preampref <= 4.0):
            raise ValueError("'preampref' must be between 0.0 and 4.0.")

        super().__init__(
            qe=qe, eta=eta, sv=sv, amp=amp, a1=a1, a2=a2, fwc=fwc, vg=vg, dt=dt
        )
        self._cutoff = cutoff
        self._vbiaspower = vbiaspower
        self._dsub = dsub
        self._vreset = vreset
        self._biasgate = biasgate
        self._preampref = preampref

    @property
    def cutoff(self) -> float:
        """Get Cutoff wavelength."""
        return self._cutoff

    @cutoff.setter
    def cutoff(self, value: float) -> None:
        """Set Cutoff wavelength."""
        if not (1.7 <= value <= 15.0):
            raise ValueError("'cutoff' must be between 1.7 and 15.0.")

        self._cutoff = value

    @property
    def vbiaspower(self) -> float:
        """Get VBIASPOWER."""
        return self._vbiaspower

    @vbiaspower.setter
    def vbiaspower(self, value: float) -> None:
        """Set VBIASPOWER."""
        if not (0.0 <= value <= 3.4):
            raise ValueError("'vbiaspower' must be between 0.0 and 3.4.")

        self._vbiaspower = value

    @property
    def dsub(self) -> float:
        """Get DSUB."""
        return self._dsub

    @dsub.setter
    def dsub(self, value: float) -> None:
        """Set DSUB."""
        if not (0.3 <= value <= 1.0):
            raise ValueError("'dsub' must be between 0.3 and 1.0.")

        self._dsub = value

    @property
    def vreset(self) -> float:
        """Get VREST."""
        return self._vreset

    @vreset.setter
    def vreset(self, value: float) -> None:
        """Set VRESET."""
        if not (0.0 <= value <= 0.3):
            raise ValueError("'vreset' must be between 0.0 and 0.3.")

        self._vreset = value

    @property
    def biasgate(self) -> float:
        """Get BIASGATE."""
        return self._biasgate

    @biasgate.setter
    def biasgate(self, value: float) -> None:
        """Set BIASGATE."""
        if not (1.8 <= value <= 2.6):
            raise ValueError("'biasgate' must be between 1.8 and 2.6.")

        self._biasgate = value

    @property
    def preampref(self) -> float:
        """Get PREAMPREF."""
        return self._preampref

    @preampref.setter
    def preampref(self, value: float) -> None:
        """Set PREAMPREF."""
        if not (0.0 <= value <= 4.0):
            raise ValueError("'preampref' must be between 0.0 and 4.0.")

        self._preampref = value



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Phase class."""

from typing import TYPE_CHECKING

import numpy as np
from astropy.units import cds

from pyxel.data_structure import Array

if TYPE_CHECKING:
    from pyxel.detectors import Geometry

cds.enable()


class Phase(Array):
    """Phase class.

    Accepted array types: np.float16, np.float32 and np.float64.
    """

    EXP_TYPE = float
    TYPE_LIST = (np.float16, np.float32, np.float64)

    def __init__(self, geo: "Geometry"):
        new_array = np.zeros((geo.row, geo.col), dtype=self.EXP_TYPE)

        super().__init__(new_array)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2021.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
"""Notebook functions."""

# flake8: noqa
from pyxel.notebook.jupyxel import (
    display_config,
    display_dict,
    display_model,
    change_modelparam,
    display_array,
    display_detector,
    display_persist,
)
from pyxel.notebook.html_representation import display_html

from pyxel.notebook.calibration import (
    display_calibration_inputs,
    display_simulated,
    display_evolution,
    champion_heatmap,
    optimal_parameters,
)

import holoviews as hv

hv.extension("bokeh")



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel general particle class to track particles like photon, electrons, holes."""

import typing as t

import numpy as np
import pandas as pd

from pyxel.util.memory import get_size

# from astropy.units import cds
# cds.enable()


class Particle:
    """Class defining and storing information of all particles with their position, velocity, energy, etc."""

    def __init__(self):
        # TODO: This should be a class variable
        self.EMPTY_FRAME = pd.DataFrame()  # type: pd.DataFrame
        self.frame = pd.DataFrame()  # type: pd.DataFrame

        self._numbytes = 0

    def get_values(self, quantity: str, id_list: t.Optional[list] = None) -> np.ndarray:
        """Get quantity values of particles defined with id_list. By default it returns values of all particles.

        :param quantity: name of quantity: ``number``, ``energy``, ``position_ver``, ``velocity_hor``, etc.
        :param id_list: list of particle ids: ``[0, 12, 321]``
        :return: array
        """
        if id_list:
            df = self.frame.query("index in %s" % id_list)  # type: pd.DataFrame
        else:
            df = self.frame

        array = df[quantity].values  # type: np.ndarray

        return array

    def set_values(
        self, quantity: str, new_value_list: list, id_list: t.Optional[list] = None
    ) -> None:
        """Update quantity values of particles defined with id_list. By default it updates all.

        :param quantity: name of quantity: ``number``, ``energy``, ``position_ver``, ``velocity_hor``, etc.
        :param new_value_list: list of values ``[1.12, 2.23, 3.65]``
        :param id_list: list of particle ids: ``[0, 12, 321]``
        """
        new_df = pd.DataFrame({quantity: new_value_list}, index=id_list)
        self.frame.update(new_df)

    def remove(self, id_list: t.Optional[list] = None) -> None:
        """Remove particles defined with id_list. By default it removes all particles from DataFrame.

        :param id_list: list of particle ids: ``[0, 12, 321]``
        """
        if id_list:
            # TODO: Check carefully if 'inplace' is needed. This could break lot of things.
            self.frame.query("index not in %s" % id_list, inplace=True)
        else:
            self.frame = self.EMPTY_FRAME.copy()

    @property
    def numbytes(self) -> int:
        """Recursively calculates object size in bytes using Pympler library.

        Returns
        -------
        int
            Size of the object in bytes.
        """
        self._numbytes = get_size(self)
        return self._numbytes



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel wrapper class for NGHXRG - Teledyne HxRG Noise Generator model."""

import logging
import typing as t
from pathlib import Path

import numpy as np
from matplotlib import pyplot as plt
from scipy import signal

from pyxel.detectors import CMOS, CMOSGeometry
from pyxel.models.charge_measurement.nghxrg.nghxrg_beta import HXRGNoise


# @pyxel.validate
# @pyxel.argument(name='', label='', units='', validate=)
def nghxrg(
    detector: CMOS,
    noise: list,
    pca0_file: t.Optional[str] = None,
    window_position: t.Optional[t.Sequence[int]] = None,
    window_size: t.Optional[t.Sequence[int]] = None,
    plot_psd: t.Optional[bool] = True,
) -> None:
    """TBW.

    :param detector: Pyxel Detector object
    :param noise:
    :param pca0_file:
    :param window_position: [x0 (columns), y0 (rows)]
    :param window_size: [x (columns), y (rows)]
    :param plot_psd:
    """
    logging.getLogger("nghxrg").setLevel(logging.WARNING)
    logging.info("")
    geo = detector.geometry  # type: CMOSGeometry
    step = 1
    if detector.is_dynamic:
        step = int(detector.time / detector.time_step)
    if window_position is None:
        window_position = [0, 0]
    if window_size is None:
        window_size = [geo.col, geo.row]
    if window_position == [0, 0] and window_size == [geo.col, geo.row]:
        window_mode = "FULL"
    else:
        window_mode = "WINDOW"

    ng = HXRGNoise(
        n_out=geo.n_output,
        time_step=step,
        nroh=geo.n_row_overhead,
        nfoh=geo.n_frame_overhead,
        reverse_scan_direction=geo.reverse_scan_direction,
        reference_pixel_border_width=geo.reference_pixel_border_width,
        pca0_file=pca0_file,
        det_size_x=geo.col,
        det_size_y=geo.row,
        wind_mode=window_mode,
        wind_x_size=window_size[0],
        wind_y_size=window_size[1],
        wind_x0=window_position[0],
        wind_y0=window_position[1],
        verbose=True,
    )  # type: HXRGNoise

    for item in noise:
        result = None  # type: t.Optional[np.ndarray]
        if "ktc_bias_noise" in item:
            # NOTE: there is no kTc or Bias noise added for first/single frame
            noise_name = "ktc_bias_noise"
            if item["ktc_bias_noise"]:
                result = ng.add_ktc_bias_noise(
                    ktc_noise=item["ktc_noise"],
                    bias_offset=item["bias_offset"],
                    bias_amp=item["bias_amp"],
                )
        elif "white_read_noise" in item:
            noise_name = "white_read_noise"
            if item["white_read_noise"]:
                result = ng.add_white_read_noise(
                    rd_noise=item["rd_noise"],
                    reference_pixel_noise_ratio=item["ref_pixel_noise_ratio"],
                )
        elif "corr_pink_noise" in item:
            noise_name = "corr_pink_noise"
            if item["corr_pink_noise"]:
                result = ng.add_corr_pink_noise(c_pink=item["c_pink"])
        elif "uncorr_pink_noise" in item:
            noise_name = "uncorr_pink_noise"
            if item["uncorr_pink_noise"]:
                result = ng.add_uncorr_pink_noise(u_pink=item["u_pink"])
        elif "acn_noise" in item:
            noise_name = "acn_noise"
            if item["acn_noise"]:
                result = ng.add_acn_noise(acn=item["acn"])
        elif "pca_zero_noise" in item:
            noise_name = "pca_zero_noise"
            if item["pca_zero_noise"]:
                result = ng.add_pca_zero_noise(pca0_amp=item["pca0_amp"])
        else:
            noise_name = ""

        if result is None:
            raise NotImplementedError

        try:
            result = ng.format_result(result)
            if result.any():
                if plot_psd:
                    display_noisepsd(
                        result,
                        noise_name=noise_name,
                        nb_output=geo.n_output,
                        dimensions=(geo.col, geo.row),
                        path=detector.output_dir,
                        step=step,
                    )
                if window_mode == "FULL":
                    detector.pixel.array += result
                elif window_mode == "WINDOW":
                    start_y_idx = window_position[1]
                    end_y_idx = window_position[1] + window_size[1]

                    start_x_idx = window_position[0]
                    end_x_idx = window_position[0] + window_size[0]

                    detector.pixel.array[
                        start_y_idx:end_y_idx, start_x_idx:end_x_idx
                    ] += result
        except TypeError:
            pass


# TODO: This generates plot. It should be in class `Output`
def display_noisepsd(
    array: np.ndarray,
    nb_output: int,
    dimensions: t.Tuple[int, int],
    noise_name: str,
    path: Path,
    step: int,
    mode: str = "plot",
) -> t.Tuple[t.Any, np.ndarray]:
    """Display noise PSD from the generated FITS file."""
    # Conversion gain
    conversion_gain = 1
    data_corr = array * conversion_gain

    """
    For power spectra, need to be careful of readout directions

    For the periodogram, using Welch's method to smooth the PSD
    > Divide data in N segment of length nperseg which overlaps at nperseg/2
    (noverlap argument)
    nperseg high means less averaging for the PSD but more points
    """
    if nb_output <= 1:
        raise ValueError("Parameter 'nb_output' must be >= 1.")

    # Should be in the simulated data header
    dimension = dimensions[0]  # type: int
    pix_p_output = dimension ** 2 / nb_output  # Number of pixels per output
    nbcols_p_channel = dimension / nb_output  # Number of columns per channel
    nperseg = int(pix_p_output / 10.0)  # Length of segments for Welch's method
    read_freq = 100000  # Frame rate [Hz]

    # Initializing table of nb_outputs periodogram +1 for dimensions
    pxx_outputs = np.zeros((nb_output, int(nperseg / 2) + 1))

    # For each output
    for i in np.arange(nb_output):
        # If i odd, flatten data since its the good reading direction
        if i % 2 == 0:
            start_x_idx = int(i * nbcols_p_channel)
            end_x_idx = int((i + 1) * nbcols_p_channel)

            output_data = data_corr[:, start_x_idx:end_x_idx].flatten()
        # Else, flip it left/right and then flatten it
        else:
            start_x_idx = int(i * nbcols_p_channel)
            end_x_idx = int((i + 1) * nbcols_p_channel)

            output_data = np.fliplr(data_corr[:, start_x_idx:end_x_idx])
            output_data.flatten()
        # output data without flipping
        # output_data = data_corr[:,int(i*(nbcols_p_channel)):
        #                         int((i+1)*(nbcols_p_channel))].flatten()

        # print(output_data, read_freq, nperseg)
        # Add periodogram to the previously initialized array
        f_vect, pxx_outputs[i] = signal.welch(output_data, read_freq, nperseg=nperseg)

    # For the detector
    # detector_data = data_corr.flatten()
    # Add periodogram to the previously initialized array
    # test, Pxx_detector = signal.welch(detector_data,
    #                                   read_freq,
    #                                   nperseg=nperseg)

    if mode == "plot":
        fig, ax1 = plt.subplots(1, 1, figsize=(8, 8))
        fig.canvas.set_window_title("Power Spectral Density")
        fig.suptitle(
            noise_name
            + " Power Spectral Density\n"
            + "Welch seg. length / Nb pixel output: "
            + str("{:1.2f}".format(nperseg / pix_p_output))
        )
        ax1.plot(f_vect, np.mean(pxx_outputs, axis=0), ".-", ms=3, alpha=0.5, zorder=32)
        for _, ax in enumerate([ax1]):
            ax.set_xlim([1, 1e5])
            ax.set_xscale("log")
            ax.set_yscale("log")
            ax.set_xlabel("Frequency [Hz]")
            ax.set_ylabel("PSD [e-${}^2$/Hz]")
            ax.grid(True, alpha=0.4)

        filename = path.joinpath(
            "nghxrg_" + noise_name + "_" + str(step) + ".png"
        )  # type: Path
        plt.savefig(filename, dpi=300)
        plt.close()

    result = np.asarray(np.mean(pxx_outputs, axis=0))  # type: np.ndarray
    return f_vect, result



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel detector simulation framework.

Pyxel is a detector simulation framework, that can simulate a variety of
detector effects (e.g., cosmic rays, radiation-induced CTI in CCDs, persistence
in MCT, charge diffusion, crosshatches, noises, crosstalk etc.) on a given image.
"""
import argparse
import logging
import sys
import time
import typing as t
from pathlib import Path

import dask
import numpy as np
import pandas as pd
import xarray as xr
from matplotlib import pyplot as plt

from pyxel import __version__ as version
from pyxel import outputs
from pyxel.calibration import Calibration, CalibrationResult
from pyxel.configuration import Configuration, load, save
from pyxel.detectors import CCD, CMOS, MKID, Detector
from pyxel.exposure import Exposure
from pyxel.observation import Observation, ObservationResult
from pyxel.pipelines import DetectionPipeline, Processor
from pyxel.util import create_model, download_examples

if t.TYPE_CHECKING:
    from .outputs import CalibrationOutputs, ExposureOutputs, ObservationOutputs


def exposure_mode(
    exposure: "Exposure",
    detector: Detector,
    pipeline: "DetectionPipeline",
) -> xr.Dataset:
    """Run an 'exposure' pipeline.

    Parameters
    ----------
    exposure
    detector
    pipeline

    Returns
    -------
    None
    """

    logging.info("Mode: Exposure")

    exposure_outputs = exposure.outputs  # type: ExposureOutputs

    detector.set_output_dir(exposure_outputs.output_dir)  # TODO: Remove this

    processor = Processor(detector=detector, pipeline=pipeline)

    result = exposure.run_exposure(processor=processor)

    if exposure_outputs.save_exposure_data:
        exposure_outputs.save_exposure_outputs(dataset=result)

    return result


def observation_mode(
    observation: "Observation",
    detector: Detector,
    pipeline: "DetectionPipeline",
    with_dask: bool = False,
) -> "ObservationResult":
    """Run an 'observation' pipeline.

    Parameters
    ----------
    observation: Observation
    detector: Detector
    pipeline: Pipeline
    with_dask: bool

    Returns
    -------
    result: ObservationResult
    """
    logging.info("Mode: Parametric")

    observation_outputs = observation.outputs  # type: ObservationOutputs
    detector.set_output_dir(observation_outputs.output_dir)  # TODO: Remove this

    # TODO: This should be done during initializing of object `Configuration`
    # parametric_outputs.params_func(parametric)

    processor = Processor(detector=detector, pipeline=pipeline)

    result = observation.run_parametric(processor=processor)

    if observation_outputs.save_observation_data:
        observation_outputs.save_observation_datasets(
            result=result, mode=observation.parameter_mode
        )

    return result


def calibration_mode(
    calibration: "Calibration",
    detector: Detector,
    pipeline: "DetectionPipeline",
    compute_and_save: bool = True,
) -> t.Tuple[xr.Dataset, pd.DataFrame, pd.DataFrame, t.Sequence]:
    """Run a 'calibration' pipeline.

    Parameters
    ----------
    calibration: Calibration
    detector: Detector
    pipeline: DetectionPipeline
    compute_and_save: bool

    Returns
    -------
    tuple
    """
    logging.info("Mode: Calibration")

    calibration_outputs = calibration.outputs  # type: CalibrationOutputs
    detector.set_output_dir(calibration_outputs.output_dir)  # TODO: Remove this

    processor = Processor(detector=detector, pipeline=pipeline)

    ds_results, df_processors, df_all_logs = calibration.run_calibration(
        processor=processor, output_dir=calibration_outputs.output_dir
    )

    # TODO: Save the processors from 'df_processors'
    # TODO: Generate plots from 'ds_results'

    # TODO: Do something with 'df_all_logs' ?

    # TODO: create 'output' object with .calibration_outputs
    # TODO: use 'fitting.get_simulated_data' ==> np.ndarray

    # geometry = processor.detector.geometry
    # calibration.post_processing(
    #     champions=champions,
    #     output=calibration_outputs,
    #     row=geometry.row,
    #     col=geometry.col,
    # )
    filenames = calibration.post_processing(
        ds=ds_results, df_processors=df_processors, output=calibration_outputs
    )

    if compute_and_save:

        computed_ds, df_processors, df_logs, filenames = dask.compute(
            ds_results, df_processors, df_all_logs, filenames
        )

        if calibration_outputs.save_calibration_data:
            calibration_outputs.save_calibration_outputs(
                dataset=computed_ds, logs=df_logs
            )
            print(f"Saved calibration outputs to {calibration_outputs.output_dir}")

        result = CalibrationResult(
            dataset=computed_ds,
            processors=df_processors,
            logs=df_logs,
            filenames=filenames,
        )

    else:
        result = CalibrationResult(
            dataset=ds_results,
            processors=df_processors,
            logs=df_all_logs,
            filenames=filenames,
        )

    return result


def output_directory(configuration: Configuration) -> Path:
    """Return the output directory from the configuration.

    Parameters
    ----------
    configuration

    Returns
    -------
    output_dir
    """
    if isinstance(configuration.exposure, Exposure):
        output_dir = configuration.exposure.outputs.output_dir
    elif isinstance(configuration.calibration, Calibration):
        output_dir = configuration.calibration.outputs.output_dir
    elif isinstance(configuration.observation, Observation):
        output_dir = configuration.observation.outputs.output_dir
    else:
        raise (ValueError("Outputs not initialized."))
    return output_dir


def run(input_filename: str, random_seed: t.Optional[int] = None) -> None:
    """TBW.

    Parameters
    ----------
    input_filename
    random_seed
    """
    logging.info("Pyxel version %s", version)
    logging.info("Pipeline started.")

    start_time = time.time()
    if random_seed:
        np.random.seed(random_seed)

    configuration = load(
        Path(input_filename).expanduser().resolve()
    )  # type: Configuration

    output_dir = output_directory(configuration)

    save(input_filename=input_filename, output_dir=output_dir)

    pipeline = configuration.pipeline  # type: DetectionPipeline

    if isinstance(configuration.ccd_detector, CCD):
        detector = configuration.ccd_detector  # type: t.Union[CCD, CMOS, MKID]
    elif isinstance(configuration.cmos_detector, CMOS):
        detector = configuration.cmos_detector
    elif isinstance(configuration.mkid_detector, MKID):
        detector = configuration.mkid_detector
    else:
        raise NotImplementedError("Detector is not defined in YAML config. file!")

    if isinstance(configuration.exposure, Exposure):
        exposure = configuration.exposure  # type: Exposure
        exposure_mode(exposure=exposure, detector=detector, pipeline=pipeline)

    elif isinstance(configuration.calibration, Calibration):

        calibration = configuration.calibration  # type: Calibration
        _ = calibration_mode(
            calibration=calibration, detector=detector, pipeline=pipeline
        )

    elif isinstance(configuration.observation, Observation):
        observation = configuration.observation  # type: Observation
        observation_mode(observation=observation, detector=detector, pipeline=pipeline)

    else:
        raise NotImplementedError("Please provide a valid simulation mode !")

    logging.info("Pipeline completed.")
    logging.info("Running time: %.3f seconds" % (time.time() - start_time))
    # Closing the logger in order to be able to move the file in the output dir
    logging.shutdown()
    outputs.save_log_file(output_dir)
    plt.close()


# TODO: Use library 'click' instead of 'parser' ? See issue #62
#       Add an option to display colors ? (very optional)
def main() -> None:
    """Define the argument parser and run Pyxel."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter, description=__doc__
    )

    parser.add_argument(
        "-v",
        "--verbosity",
        action="count",
        default=0,
        help="Increase output verbosity (-v/-vv/-vvv)",
    )

    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version="Pyxel, version {version}".format(version=version),
    )

    parser.add_argument(
        "-c",
        "--config",
        type=str,
        help="Configuration file to load (YAML)",
    )

    parser.add_argument("-s", "--seed", type=int, help="Random seed for the framework")

    parser.add_argument(
        "--download-examples",
        nargs="?",
        const="pyxel-examples",
        help="Install examples to the specified directory, default is /pyxel-examples.",
    )

    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Force flag for saving the examples.",
    )

    parser.add_argument(
        "-cm",
        "--createmodel",
        type=str,
        help="""Use: -cm arg1/arg2. Create a new module in\
        pyxel/models/arg1/arg2 using a template\
        (pyxel/templates/MODELTEMPLATE.py)""",
    )

    # parser.add_argument('-g', '--gui', default=False, type=bool, help='run Graphical User Interface')
    # parser.add_argument('-p', '--port', default=9999, type=int, help='The port to run the web server on')

    opts = parser.parse_args()

    logging_level = [logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG][
        min(opts.verbosity, 3)
    ]
    log_format = (
        "%(asctime)s - %(name)s - %(threadName)30s - %(funcName)30s \t %(message)s"
    )
    logging.basicConfig(
        filename="pyxel.log",
        level=logging_level,
        format=log_format,
        datefmt="%d-%m-%Y %H:%M:%S",
    )
    # If user wants the log in stdout AND in file, use the three lines below
    stream_stdout = logging.StreamHandler(sys.stdout)
    stream_stdout.setFormatter(logging.Formatter(log_format))
    logging.getLogger().addHandler(stream_stdout)

    if opts.config:
        run(input_filename=opts.config, random_seed=opts.seed)
    elif opts.download_examples:
        download_examples(foldername=opts.download_examples, force=opts.force)
    elif opts.createmodel:
        create_model(newmodel=opts.createmodel)
    else:
        print("Define a YAML configuration file!")


if __name__ == "__main__":
    main()



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""
import typing as t

import numpy as np

from pyxel.util.memory import get_size


# TODO: 'vg' should be the full volume and not the half
class Characteristics:
    """Characteristical attributes of the detector."""

    def __init__(
        self,
        qe: float = 0.0,  # unit: NA
        eta: float = 0.0,  # unit: electron/photon
        sv: float = 0.0,  # unit: volt/electron
        amp: float = 0.0,  # unit: V/V
        a1: float = 0.0,  # unit: V/V
        a2: int = 0,  # unit: adu/V
        fwc: int = 0,  # unit: electron
        vg: float = 0.0,  # unit: cm^2
        dt: float = 0.0,  # unit: s
    ):
        """Create an instance of `Characteristics`.

        Parameters
        ----------
        qe: float
            Quantum efficiency.
        eta: float
            Quantum yield. Unit: e-/photon
        sv: float
            Sensitivity of charge readout. Unit: V/e-
        amp: float
            Gain of output amplifier. Unit: V/V
        a1: float
            Gain of the signal processor. Unit: V/V
        a2: int
            Gain of the Analog-Digital Converter. Unit: ADU/V
        fwc: int
            Full well capacity. Unit: e-
        vg: float
            Half pixel volume charge can occupy. Unit: cm^2
        dt: float
            Pixel dwell time. Unit: s
        """
        if not (0.0 <= qe <= 1.0):
            raise ValueError("'qe' must be between 0.0 and 1.0.")

        if not (0.0 <= eta <= 1.0):
            raise ValueError("'eta' must be between 0.0 and 1.0.")

        if not (0.0 <= sv <= 100.0):
            raise ValueError("'sv' must be between 0.0 and 100.0.")

        if not (0.0 <= amp <= 100.0):
            raise ValueError("'amp' must be between 0.0 and 100.0.")

        if not (0.0 <= a1 <= 100.0):
            raise ValueError("'a1' must be between 0.0 and 100.0.")

        if a2 not in range(65537):
            raise ValueError("'a2' must be between 0 and 65536.")

        if fwc not in range(10000001):
            raise ValueError("'fwc' must be between 0 and 1e+7.")

        if not (0.0 <= vg <= 1.0):
            raise ValueError("'vg' must be between 0.0 and 1.0.")

        if not (0.0 <= dt <= 10.0):
            raise ValueError("'dt' must be between 0.0 and 10.0.")

        self._qe = qe  # type: t.Union[float, np.ndarray]
        self._eta = eta
        self._sv = sv
        self._amp = amp
        self._a1 = a1
        self._a2 = a2
        self._fwc = fwc
        self._vg = vg
        self._dt = dt

        self._numbytes = 0

    @property
    def qe(self) -> t.Union[float, np.ndarray]:
        """Get Quantum efficiency."""
        return self._qe

    @qe.setter
    def qe(self, value: t.Union[float, np.ndarray]) -> None:
        """Set Quantum efficiency."""
        if not (0.0 <= np.min(value)) and (np.max(value) <= 1.0):
            raise ValueError("'QE' values must be between 0.0 and 1.0.")

        self._qe = value

    @property
    def eta(self) -> float:
        """Get Quantum yield."""
        return self._eta

    @eta.setter
    def eta(self, value: float) -> None:
        """Set Quantum yield."""
        if not (0.0 <= value <= 1.0):
            raise ValueError("'eta' must be between 0.0 and 1.0.")

        self._eta = value

    @property
    def sv(self) -> float:
        """Get Sensitivity of charge readout."""
        return self._sv

    @sv.setter
    def sv(self, value: float) -> None:
        """Set Sensitivity of charge readout."""
        if not (0.0 <= value <= 100.0):
            raise ValueError("'sv' must be between 0.0 and 100.0.")
        self._sv = value

    @property
    def amp(self) -> float:
        """Get Gain of output amplifier."""
        return self._amp

    @amp.setter
    def amp(self, value: float) -> None:
        """Set Gain of output amplifier."""
        if not (0.0 <= value <= 100.0):
            raise ValueError("'amp' must be between 0.0 and 100.0.")

        self._amp = value

    @property
    def a1(self) -> float:
        """Get Gain of the signal processor."""
        return self._a1

    @a1.setter
    def a1(self, value: float) -> None:
        """Set Gain of the signal processor."""
        if not (0.0 <= value <= 100.0):
            raise ValueError("'a1' must be between 0.0 and 100.0.")

        self._a1 = value

    @property
    def a2(self) -> int:
        """Get Gain of the Analog-Digital Converter."""
        return self._a2

    @a2.setter
    def a2(self, value: int) -> None:
        """Set Gain of the Analog-Digital Converter."""
        if value not in range(65537):
            raise ValueError("'a2' must be between 0 and 65536.")

        self._a2 = value

    @property
    def fwc(self) -> int:
        """Get Full well capacity."""
        return self._fwc

    @fwc.setter
    def fwc(self, value: int) -> None:
        """Set Full well capacity."""
        if value not in range(10000001):
            raise ValueError("'fwc' must be between 0 and 1e+7.")

        self._fwc = value

    @property
    def vg(self) -> float:
        """Get Half pixel volume charge can occupy."""
        return self._vg

    @vg.setter
    def vg(self, value: float) -> None:
        """Set Half pixel volume charge can occupy."""
        if not (0.0 <= value <= 1.0):
            raise ValueError("'vg' must be between 0.0 and 1.0.")

    @property
    def dt(self) -> float:
        """Get Pixel dwell time."""
        return self._dt

    @dt.setter
    def dt(self, value: float) -> None:
        """Set Pixel dwell time."""
        if not (0.0 <= value <= 10.0):
            raise ValueError("'dt' must be between 0.0 and 10.0.")

        self._dt = value

    @property
    def numbytes(self) -> int:
        """Recursively calculates object size in bytes using Pympler library.

        Returns
        -------
        int
            Size of the object in bytes.
        """
        self._numbytes = get_size(self)
        return self._numbytes



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel full well models."""
import logging

from pyxel.detectors import Detector


# @pyxel.validate
# @pyxel.argument(name='', label='', units='', validate=)
def simple_full_well(detector: Detector) -> None:
    """Limiting the amount of charge in pixel due to full well capacity."""
    logging.info("")
    fwc = detector.characteristics.fwc
    if fwc is None:
        raise ValueError("Full Well Capacity is not defined")
    charge_array = detector.pixel.array
    charge_array[charge_array > fwc] = fwc
    detector.pixel.array = charge_array



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""
from pyxel.detectors import Geometry


class CMOSGeometry(Geometry):
    """Geometrical attributes of a CMOS-based detector."""

    def __init__(
        self,
        # Parameters for Geometry
        row: int = 0,
        col: int = 0,
        total_thickness: float = 0.0,
        pixel_vert_size: float = 0.0,
        pixel_horz_size: float = 0.0,
        # Parameters for CMOS Geometry
        n_output: int = 1,
        n_row_overhead: int = 0,
        n_frame_overhead: int = 0,
        reverse_scan_direction: bool = False,
        reference_pixel_border_width: int = 4,
    ):
        """Create a new instance of `CMOSGeometry`.

        Parameters
        ----------
        n_output: int
            Number of detector outputs.
        n_row_overhead: int
            New row overhead in pixel.
            This allows for a short wait at the end of a row before starting the next row.
        n_frame_overhead: int
            New frame overhead in rows.
            This allows for a short wait at the end of a frame before starting the next frame.
        reverse_scan_direction: bool
            Set this True to reverse the fast scanner readout directions.
            This capability was added to support Teledyne’s programmable fast scan readout directions.
            The default setting (False) corresponds to what HxRG detectors default to upon power up.
        reference_pixel_border_width: int
            Width of reference pixel border around image area.
        """
        if n_output not in range(33):
            raise ValueError("'n_output' must be between 0 and 32.")

        if n_row_overhead not in range(101):
            raise ValueError("'n_row_overhead' must be between 0 and 100.")

        if n_frame_overhead not in range(101):
            raise ValueError("'n_frame_overhead' must be between 0 and 100.")

        if reference_pixel_border_width not in range(33):
            raise ValueError("'reference_pixel_border_width' must be between 0 and 32.")

        super().__init__(
            row=row,
            col=col,
            total_thickness=total_thickness,
            pixel_vert_size=pixel_vert_size,
            pixel_horz_size=pixel_horz_size,
        )

        self._n_output = n_output
        self._n_row_overhead = n_row_overhead
        self._n_frame_overhead = n_frame_overhead
        self._reverse_scan_direction = reverse_scan_direction
        self._reference_pixel_border_width = reference_pixel_border_width

    @property
    def n_output(self) -> int:
        """Get Number of detector outputs."""
        return self._n_output

    @n_output.setter
    def n_output(self, value: int) -> None:
        """Set Number of detector outputs."""
        if value not in range(33):
            raise ValueError("'n_output' must be between 0 and 32.")

        self._n_output = value

    @property
    def n_row_overhead(self) -> int:
        """Get Number of detector outputs."""
        return self._n_row_overhead

    @n_row_overhead.setter
    def n_row_overhead(self, value: int) -> None:
        """Set Number of detector outputs."""
        if value not in range(101):
            raise ValueError("'n_row_overhead' must be between 0 and 100.")

        self._n_row_overhead = value

    @property
    def n_frame_overhead(self) -> int:
        """Get New frame overhead in rows."""
        return self._n_frame_overhead

    @n_frame_overhead.setter
    def n_frame_overhead(self, value: int) -> None:
        """Set New frame overhead in rows."""
        if value not in range(101):
            raise ValueError("'n_frame_overhead' must be between 0 and 100.")

        self._n_frame_overhead = value

    @property
    def reverse_scan_direction(self) -> bool:
        """Get reverse scan direction."""
        return self._reverse_scan_direction

    @reverse_scan_direction.setter
    def reverse_scan_direction(self, value: bool) -> None:
        """Set reverse scan direction."""
        self._reverse_scan_direction = value

    @property
    def reference_pixel_border_width(self) -> int:
        """Get Number of detector outputs."""
        return self._reference_pixel_border_width

    @reference_pixel_border_width.setter
    def reference_pixel_border_width(self, value: int) -> None:
        """Set Number of detector outputs."""
        if value not in range(33):
            raise ValueError("'reference_pixel_border_width' must be between 0 and 32.")

        self._reference_pixel_border_width = value



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2021.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#

import pytest

from pyxel.pipelines import Arguments


@pytest.fixture
def arguments() -> Arguments:
    """Create a valid `Arguments` instance."""
    args = {"one": 1, "two": 2, "three": 3}
    return Arguments(args)


def test_type_model_function(arguments: Arguments):
    assert isinstance(arguments, Arguments)


def test_getitem(arguments: Arguments):
    assert arguments["one"] == 1


def test_getitem_key_error(arguments: Arguments):
    with pytest.raises(KeyError):
        _ = arguments["four"]


def test_setitem(arguments: Arguments):
    arguments["two"] = 0
    assert arguments["two"] == 0


def test_setitem_key_error(arguments: Arguments):
    with pytest.raises(KeyError):
        arguments["four"] = 4


def test_getattr(arguments: Arguments):
    assert arguments.one == 1
    assert arguments._arguments == {"one": 1, "two": 2, "three": 3}


def test_getattr_attribute_error(arguments: Arguments):
    with pytest.raises(AttributeError):
        _ = arguments.four


def test_setattr(arguments: Arguments):
    arguments.two = 0
    assert arguments.two == 0


def test_setattr_attribute_error(arguments: Arguments):
    with pytest.raises(AttributeError):
        arguments.four = 4


def test_len(arguments: Arguments):
    assert len(arguments) == 3



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel photon generator models."""
import typing as t
from pathlib import Path

import numpy as np

from pyxel import inputs
from pyxel.data_structure import Photon
from pyxel.detectors import Detector


# TODO: Fix this
# @validators.validate
# @config.argument(name='image_file', label='fits file', validate=check_path)
# @config.argument(name='fit_image_to_det', label='fitting image to detector', validate=check_type(bool))
# @config.argument(name='convert_to_photons', label='convert ADU values to photon numbers', validate=check_type(bool))
def load_image(
    detector: Detector,
    image_file: str,
    fit_image_to_det: bool = False,
    position: t.Tuple[int, int] = (0, 0),  # TODO Too many arguments
    convert_to_photons: bool = False,
) -> None:
    r"""Load FITS file as a numpy array and add to the detector as input image.

    :param detector: Pyxel Detector object
    :param image_file: path to FITS image file
    :param fit_image_to_det: fitting image to detector shape (Geometry.row, Geometry.col)
    :param position: indices of starting row and column, used when fitting image to detector
    :param convert_to_photons: if ``True``, the model converts the values of loaded image array from ADU to
        photon numbers for each pixel using the Photon Transfer Function:
        :math:`PTF = QE \cdot \eta \cdot S_{v} \cdot amp \cdot a_{1} \cdot a_{2}`
    """
    filename = Path(image_file).expanduser().resolve()

    if not Path(filename).exists():
        raise FileNotFoundError(f"Image file '{filename}' does not exist !")

    image = inputs.load_image(filename)  # type: np.ndarray

    if fit_image_to_det:
        geo = detector.geometry
        position_y, position_x = position

        image = image[
            slice(position_y, position_y + geo.row),
            slice(position_x, position_x + geo.col),
        ]

    detector.input_image = image
    photon_array = image

    if convert_to_photons:
        cht = detector.characteristics
        photon_array = photon_array / (
            cht.qe * cht.eta * cht.sv * cht.amp * cht.a1 * cht.a2
        )

    detector.photon = Photon(photon_array)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""

from pyxel.util.memory import get_size


class Environment:
    """Environmental attributes of the detector."""

    def __init__(
        self,
        temperature: float = 273.15,
        total_ionising_dose: float = 0.0,
        total_non_ionising_dose: float = 0.0,
    ):
        """Create a new instance of `Environment`.

        Parameters
        ----------
        temperature: float
            Temperature of the detector. Unit: K
        total_ionising_dose: float
            Total Ionising Dose (TID) of the detector. Unit: MeV/g
        total_non_ionising_dose: float
            Total Non-Ionising Dose (TNID) of the detector. Unit: MeV/g
        """
        if not (0.0 <= temperature <= 1000.0):
            raise ValueError("'temperature' must be between 0.0 and 1000.0.")

        if not (0.0 <= total_ionising_dose <= 1e15):
            raise ValueError("'total_ionising_dose' must be between 0.0 and 1e15.")

        if not (0.0 <= total_non_ionising_dose <= 1e15):
            raise ValueError("'total_non_ionising_dose' must be between 0.0 and 1e15.")

        self._temperature = temperature
        self._total_ionising_dose = total_ionising_dose
        self._total_non_ionising_dose = total_non_ionising_dose

        self._numbytes = 0

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__  # type: str
        return (
            f"{cls_name}(temperature={self._temperature!r}, "
            f"total_ionising_dose={self._total_ionising_dose!r}, "
            f"total_non_ionising_dose={self._total_non_ionising_dose!r})"
        )

    @property
    def temperature(self) -> float:
        """Get Temperature of the detector."""
        return self._temperature

    @temperature.setter
    def temperature(self, value: float) -> None:
        """Set Temperature of the detector."""
        if not (0.0 <= value <= 1000.0):
            raise ValueError("'temperature' must be between 0.0 and 1000.0.")

        self._temperature = value

    @property
    def total_ionising_dose(self) -> float:
        """Get Total Ionising Dose (TID) of the detector."""
        return self._total_ionising_dose

    @total_ionising_dose.setter
    def total_ionising_dose(self, value: float) -> None:
        """Set Total Ionising Dose (TID) of the detector."""
        if not (0.0 <= value <= 1e15):
            raise ValueError("'total_ionising_dose' must be between 0.0 and 1e15.")

        self._total_ionising_dose = value

    @property
    def total_non_ionising_dose(self) -> float:
        """Get Total Ionising Dose (TID) of the detector."""
        return self._total_non_ionising_dose

    @total_non_ionising_dose.setter
    def total_non_ionising_dose(self, value: float) -> None:
        """Set Total Ionising Dose (TID) of the detector."""
        if not (0.0 <= value <= 1e15):
            raise ValueError("'total_non_ionising_dose' must be between 0.0 and 1e15.")

        self._total_non_ionising_dose = value

    @property
    def numbytes(self) -> int:
        """Recursively calculates object size in bytes using Pympler library.

        Returns
        -------
        int
            Size of the object in bytes.
        """
        self._numbytes = get_size(self)
        return self._numbytes



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Subpackage used to display the versions of all dependencies."""
import importlib
import locale
import os
import platform
import struct
import sys
import typing as t

from ._version import get_versions

__all__ = ["show_versions"]


def get_system_info() -> t.Mapping[str, t.Any]:
    """Get extra informations."""
    # Get git commit hash
    commit = get_versions()["full-revisionid"]  # type: t.Optional[str]
    version = get_versions()["version"]  # type: str

    size_integer = struct.calcsize("P")  # type: int
    language_code, encoding = locale.getlocale()

    return {
        "commit": commit,
        "version": version,
        "python": sys.version,
        "python-bits": size_integer * 8,
        "OS": platform.system(),
        "OS-release": platform.release(),
        "machine": platform.machine(),
        "processor": platform.processor(),
        "byteorder": sys.byteorder,
        "LC_ALL": os.environ.get("LC_ALL"),
        "LANG": os.environ.get("LANG"),
        "LOCALE": f"{language_code}.{encoding}",
    }


def show_versions():
    """Display the versions of Pyxel and its dependencies.

    Examples
    --------
    >>> import pyxel
    >>> pyxel.show_versions()
    INSTALLED VERSIONS
    ------------------
    commit     : cfb7ce2886d24f884699aafd1ad7dc6f8615252d
    version    : 0.6+18.gcfb7ce2.dirty
    python     : 3.8.5 | packaged by conda-forge | (default, Aug 29 2020, 01:18:42)
    [Clang 10.0.1 ]
    python-bits: 64
    OS         : Darwin
    OS-release : 19.5.0
    machine    : x86_64
    processor  : i386
    byteorder  : little
    LC_ALL     : None
    LANG       : None
    LOCALE     : None.UTF-8
    pyxel      : 0.6+18.gcfb7ce2.dirty
    astropy    : 4.0.1.post1
    dask       : 2.25.0
    distributed: 2.25.0
    h5py       : 2.10.0
    ipywidgets : 7.5.1
    jupyter    : installed
    jupyterlab : 2.2.7
    matplotlib : 3.3.1
    numba      : 0.51.2
    numpy      : 1.19.1
    pandas     : 1.1.2
    poppy      : 0.9.1
    pygmo      : 2.15.0
    scipy      : 1.5.2
    setuptools : 49.6.0.post20200814
    pip        : 20.2.3
    conda      : 4.8.4
    black      : 20.8b1
    flake8     : 3.8.3
    isort      : 5.5.2
    mypy       : installed
    pytest     : 6.0.1
    sphinx     : None
    """
    system_info = get_system_info()  # type: t.Mapping[str, t.Any]

    dependencies_lst = (
        "pyxel",
        # required
        "astropy",
        "bokeh",
        "cloudpickle",
        "dask",
        "dask_jobqueue",
        "distributed",
        "fsspec",
        "h5py",
        "holoviews",
        "ipywidgets",
        "jupyter",
        "jupyterlab",
        "matplotlib",
        "notebook",
        "numba",
        "numpy",
        "pandas",
        "PIL",
        "poppy",
        "pygmo",
        "pympler",
        "scipy",
        "seaborn",
        "skimage",
        "tqdm",
        "typing-extensions",
        "yaml",
        "xarray",
        "xlrd",
        "openpyxl",
        # Install / build
        "setuptools",
        "pip",
        "conda",
        # Test
        "black",
        "blackdoc",
        "flake8",
        "isort",
        "mypy",
        "pytest",
        "tox",
        # Docs
        "sphinx",
    )  # type: t.Tuple[str, ...]

    dependencies = {}  # type: dict

    for module_name in dependencies_lst:
        try:
            # Try to get a module
            module = importlib.import_module(module_name)
        except Exception:
            dependencies[module_name] = None
        else:
            try:
                # Try to get a version
                version = module.__version__  # type: ignore
                dependencies[module_name] = version
            except Exception:
                dependencies[module_name] = "installed"

    max_length = max(map(len, [*system_info, *dependencies]))
    print("")
    print("INSTALLED VERSIONS")
    print("------------------")
    for key, value in system_info.items():
        print(f"{key:<{max_length}}: {value}")
    print("")
    for key, value in dependencies.items():
        print(f"{key:<{max_length}}: {value}")
